# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.py import (
    get_smallest_numpy_int_type,
    float_cast,
    int_cast,
)
from skyllh.core.utils.coords import (
    rotate_signal_events_on_sphere,
)
from skyllh.core.signal_generation import (
    SignalGenerationMethod,
)
from skyllh.core.source_model import (
    PointLikeSource,
)


def source_sin_dec_shift_linear(x, w, L, U):
    """Calculates the shift of the sine of the source declination, in order to
    allow the construction of the source sine declination band with
    sin(dec_src) +/- w. This shift function, S(x), is implemented as a line
    with the following points:

        S(L) = w
        S((L+U)/2) = 0
        S(U) = -w

    Parameters
    ----------
    x : 1D numpy ndarray
        The sine of the source declination for each source.
    w : float
        The half size of the sin(dec)-window.
    L : float
        The lower value of the allowed sin(dec) range.
    U : float
        The upper value of the allowed sin(dec) range.

    Returns
    -------
    S : 1D numpy ndarray
        The sin(dec) shift of the sin(dec) values of the given sources, such
        that ``sin(dec_src) + S`` is the new sin(dec) of the source, and
        ``sin(dec_src) + S +/- w`` is always within the sin(dec) range [L, U].
    """
    x = np.atleast_1d(x)

    m = -2*w/(U-L)
    b = w*(L+U)/(U-L)
    S = m*x+b

    return S


def source_sin_dec_shift_cubic(x, w, L, U):
    """Calculates the shift of the sine of the source declination, in order to
    allow the construction of the source sine declination band with
    sin(dec_src) +/- w. This shift function, S(x), is implemented as a cubic
    function with the following points:

        S(L) = w
        S((L+U)/2) = 0
        S(U) = -w

    Parameters
    ----------
    x : 1D numpy ndarray
        The sine of the source declination for each source.
    w : float
        The half size of the sin(dec)-window.
    L : float
        The lower value of the allowed sin(dec) range.
    U : float
        The upper value of the allowed sin(dec) range.

    Returns
    -------
    S : 1D numpy ndarray
        The sin(dec) shift of the sin(dec) values of the given sources, such
        that ``sin(dec_src) + S`` is the new sin(dec) of the source, and
        ``sin(dec_src) + S +/- w`` is always within the sin(dec) range [L, U].
    """
    # TODO: Make sure that this function does what it's supposed to do
    x = np.atleast_1d(x)

    m = w / (x - 0.5*(L+U))**3
    S = m * np.power(x-0.5*(L+U), 3)

    return S


class PointLikeSourceI3SignalGenerationMethod(SignalGenerationMethod):
    """This class provides a signal generation method for point-like sources
    seen in the IceCube detector.
    """

    def __init__(
        self,
        src_sin_dec_half_bandwidth=np.sin(np.radians(1)),
        src_sin_dec_shift_func=None,
        energy_range=None,
        src_batch_size=128,
        **kwargs
    ):
        """Constructs a new signal generation method instance for a point-like
        source detected with IceCube.

        Parameters
        ----------
        src_sin_dec_half_bandwidth : float
            The half-width of the sin(dec) band to take MC events from around a
            source. The default is sin(1deg), i.e. a 1deg half-bandwidth.
        src_sin_dec_shift_func : callable | None
            The function that provides the source sin(dec) shift needed for
            constructing the source declination bands from where to draw
            monte-carlo events from. If set to None, the default function
            ``source_sin_dec_shift_linear`` will be used.
        energy_range : 2-element tuple of float | None
            The energy range from which to take MC events into account for
            signal event generation.
            If set to None, the entire energy range [0, +inf] is used.
        src_batch_size : int
            The source processing batch size used for the signal event flux
            calculation.
        """
        super().__init__(
            energy_range=energy_range,
            **kwargs)

        self.src_sin_dec_half_bandwidth = src_sin_dec_half_bandwidth

        if src_sin_dec_shift_func is None:
            src_sin_dec_shift_func = source_sin_dec_shift_linear
        self.src_sin_dec_shift_func = src_sin_dec_shift_func

        self.src_batch_size = src_batch_size

    @property
    def src_sin_dec_half_bandwidth(self):
        """The half-width of the sin(dec) band to take MC events from around a
        source.
        """
        return self._src_sin_dec_half_bandwidth

    @src_sin_dec_half_bandwidth.setter
    def src_sin_dec_half_bandwidth(self, v):
        v = float_cast(
            v,
            'The src_sin_dec_half_bandwidth property must be cast-able to type '
            'float!')
        self._src_sin_dec_half_bandwidth = v

    @property
    def src_sin_dec_shift_func(self):
        """The function that provides the source sin(dec) shift needed for
        constructing the source declination bands from where to draw
        monte-carlo events from.
        """
        return self._src_sin_dec_shift_func

    @src_sin_dec_shift_func.setter
    def src_sin_dec_shift_func(self, func):
        if not callable(func):
            raise TypeError(
                'The src_sin_dec_shift_func property must be a callable '
                'object!')
        self._src_sin_dec_shift_func = func

    @property
    def src_batch_size(self):
        """The source processing batch size used for the signal event flux
        calculation.
        """
        return self._src_batch_size

    @src_batch_size.setter
    def src_batch_size(self, v):
        v = int_cast(
            v,
            'The src_batch_size property must be cast-able to type int!')
        self._src_batch_size = v

    def _get_src_dec_bands(self, src_dec, max_sin_dec_range):
        """Calculates the minimum and maximum sin(dec) values for each source
        to use with a specified maximal sin(dec) range, which should get
        determined from the available MC data itself.

        Parameters
        ----------
        src_dec : 1D ndarray
            The declination values of the sources.
        max_sin_dec_range : 2-element tuple of floats
            The maximal sin(dec) range from where MC events are available.

        Returns
        -------
        src_sin_dec_band_min : (N_sources,)-shaped 1D ndarray
            The array holding the lower value of the sin(dec) band for each
            source.
        src_sin_dec_band_max : (N_sources,)-shaped 1D ndarray
            The array holding the upper value of the sin(dec) band for each
            source.
        src_dec_band_omega : (N_sources,)-shaped 1D ndarray
            The solid angle of the declination band for each source.
        """
        # Shift the source declination in order to be able to always create the
        # declination band via src_dec +/- half-bandwidth
        src_sin_dec = np.sin(src_dec)
        src_sin_dec += self._src_sin_dec_shift_func(
            src_sin_dec, self._src_sin_dec_half_bandwidth, *max_sin_dec_range)

        src_sin_dec_band_min = src_sin_dec - self._src_sin_dec_half_bandwidth
        src_sin_dec_band_max = src_sin_dec + self._src_sin_dec_half_bandwidth

        # Calculate the solid angle of the declination band.
        src_dec_band_omega = (
            2 * np.pi * (src_sin_dec_band_max - src_sin_dec_band_min)
        )

        return (src_sin_dec_band_min, src_sin_dec_band_max, src_dec_band_omega)

    def calc_source_signal_mc_event_flux(self, data_mc, shg):
        """Calculates the signal flux of each given MC event for each source
        hypothesis of the given source hypothesis group.

        Parameters
        ----------
        data_mc : numpy record ndarray
            The numpy record array holding the MC events of a dataset.
        shg : SourceHypoGroup instance
            The source hypothesis group, which defines the list of sources, and
            their flux model.

        Returns
        -------
        ev_idx_arr : ndarray
            The (N_selected_signal_events,)-shaped 1D ndarray holding the index
            of the MC event.
        shg_src_idx_arr : ndarray
            The (N_selected_signal_events,)-shaped 1D ndarray holding the index
            of the source within the given source hypothesis group for each
            signal candidate event.
        flux_arr : ndarray
            The (N_selected_signal_events,)-shaped 1D ndarray holding the flux
            value of each signal candidate event.
        """
        indices = np.arange(
            0, len(data_mc),
            dtype=get_smallest_numpy_int_type((0, len(data_mc)))
        )
        n_sources = shg.n_sources

        # Get 1D array of source declination.
        src_dec = np.empty((n_sources,), dtype=np.float64)
        for (k, source) in enumerate(shg.source_list):
            if not isinstance(source, PointLikeSource):
                raise TypeError(
                    'The source instance must be an instance of '
                    'PointLikeSource!')
            src_dec[k] = source.dec

        data_mc_sin_true_dec = data_mc['sin_true_dec']
        data_mc_true_energy = data_mc['true_energy']

        # Calculate the source declination bands and their solid angle.
        max_sin_dec_range = (
            np.min(data_mc_sin_true_dec),
            np.max(data_mc_sin_true_dec)
        )
        (src_sin_dec_band_min, src_sin_dec_band_max, src_dec_band_omega) =\
            self._get_src_dec_bands(src_dec, max_sin_dec_range)

        # Get the flux model of this source hypo group (SHG).
        fluxmodel = shg.fluxmodel

        # Get the theoretical weights of all the sources of this SHG.
        src_weights = shg.get_source_weights()

        # Calculate conversion factor from the flux model unit into the internal
        # flux unit.
        to_internal_flux_unit =\
            fluxmodel.to_internal_flux_unit()

        # Select the events that belong to a given source.
        ev_idx_arr = np.empty(
            (0,),
            dtype=get_smallest_numpy_int_type((0, len(data_mc))))
        shg_src_idx_arr = np.empty(
            (0,),
            dtype=get_smallest_numpy_int_type((0, n_sources)))
        flux_arr = np.empty(
            (0,),
            dtype=np.float32)

        src_batch_size = self._src_batch_size
        n_batches = int(np.ceil(n_sources / src_batch_size))

        for bi in range(n_batches):
            src_start = bi*src_batch_size
            src_end = np.min([(bi+1)*src_batch_size, n_sources])
            bs = src_end - src_start

            src_slice = slice(src_start, src_end)

            # Create an event mask of shape (N_sources,N_events).
            ev_mask = np.logical_and(
                (data_mc_sin_true_dec >=
                    src_sin_dec_band_min[src_slice][:, np.newaxis]),
                (data_mc_sin_true_dec <=
                    src_sin_dec_band_max[src_slice][:, np.newaxis])
            )

            if self.energy_range is not None:
                ev_mask &= np.logical_and(
                    (data_mc_true_energy >= self.energy_range[0]),
                    (data_mc_true_energy <= self.energy_range[1])
                )

            ev_idxs = np.tile(indices, bs)[ev_mask.ravel()]
            shg_src_idxs = bi*src_batch_size + np.repeat(
                np.arange(bs),
                ev_mask.sum(axis=1)
            )
            del ev_mask

            flux = (
                fluxmodel(E=data_mc_true_energy[ev_idxs]).squeeze() *
                to_internal_flux_unit /
                src_dec_band_omega[shg_src_idxs]
            )
            if src_weights is not None:
                flux *= src_weights[shg_src_idxs]

            ev_idx_arr = np.append(ev_idx_arr, ev_idxs)
            shg_src_idx_arr = np.append(shg_src_idx_arr, shg_src_idxs)
            flux_arr = np.append(flux_arr, flux)

        return (ev_idx_arr, shg_src_idx_arr, flux_arr)

    def signal_event_post_sampling_processing(
            self,
            shg,
            shg_sig_events_meta,
            shg_sig_events,
    ):
        """Rotates the generated signal events to their source location for a
        given source hypothesis group.

        Parameters
        ----------
        shg : SourceHypoGroup instance
            The source hypothesis group instance holding the sources and their
            locations.
        shg_sig_events_meta : numpy record ndarray
            The numpy record ndarray holding meta information about the
            generated signal events for the given source hypothesis group.
            The length of this array must be the same as shg_sig_events.
            It needs to contain the following data fields:

            - 'shg_src_idx': int
                The source index within the source hypothesis group.

        shg_sig_events : numpy record ndarray
            The numpy record ndarray holding the generated signal events for
            the given source hypothesis group and in the format of the original
            MC events.

        Returns
        -------
        shg_sig_events : numpy record ndarray
            The numpy record ndarray with the processed MC signal events.
        """
        # Get the unique source indices of that source hypo group.
        shg_src_idxs = np.unique(shg_sig_events_meta['shg_src_idx'])
        # Go through each (sampled) source.
        for shg_src_idx in shg_src_idxs:
            source = shg.source_list[shg_src_idx]
            # Get the signal events of the source hypo group, that belong to the
            # source index.
            shg_src_mask = shg_sig_events_meta['shg_src_idx'] == shg_src_idx
            shg_src_sig_events = shg_sig_events[shg_src_mask]
            n_sig = len(shg_src_sig_events)

            # Rotate the signal events to the source location.
            (ra, dec) = rotate_signal_events_on_sphere(
                src_ra=np.full(n_sig, source.ra),
                src_dec=np.full(n_sig, source.dec),
                evt_true_ra=shg_src_sig_events['true_ra'],
                evt_true_dec=shg_src_sig_events['true_dec'],
                evt_reco_ra=shg_src_sig_events['ra'],
                evt_reco_dec=shg_src_sig_events['dec']
            )

            shg_src_sig_events['ra'] = ra
            shg_src_sig_events['dec'] = dec
            shg_src_sig_events['sin_dec'] = np.sin(dec)

            shg_sig_events[shg_src_mask] = shg_src_sig_events

        return shg_sig_events



# -*- coding: utf-8 -*-

"""Plotting module for core PDF ratio objects.
"""

import numpy as np
import itertools

from matplotlib.axes import Axes
from matplotlib.colors import LogNorm

from skyllh.core.py import classname
from skyllh.core.storage import DataFieldRecordArray
from skyllh.core.trialdata import TrialDataManager
from skyllh.core.pdfratio import SigOverBkgPDFRatio


class SigOverBkgPDFRatioPlotter(object):
    """Plotter class to plot a SigOverBkgPDFRatio object.
    """
    def __init__(self, tdm, pdfratio):
        """Creates a new plotter object for plotting a
        SpatialSigOverBkgPDFRatio object.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that provides the data for the
            PDF ratio evaluation.
        pdfratio : SpatialSigOverBkgPDFRatio
            The PDF ratio object to plot.
        """
        self.tdm = tdm
        self.pdfratio = pdfratio

    @property
    def pdfratio(self):
        """The PDF ratio object to plot.
        """
        return self._pdfratio

    @pdfratio.setter
    def pdfratio(self, pdfratio):
        if not isinstance(pdfratio, SigOverBkgPDFRatio):
            raise TypeError(
                'The pdfratio property must be an instance of '
                'SigOverBkgPDFRatio!')
        self._pdfratio = pdfratio

    @property
    def tdm(self):
        """The TrialDataManager that provides the data for the PDF evaluation.
        """
        return self._tdm

    @tdm.setter
    def tdm(self, obj):
        if not isinstance(obj, TrialDataManager):
            raise TypeError(
                'The tdm property must be an instance of TrialDataManager!')
        self._tdm = obj

    def plot(
            self,
            src_hypo_group_manager,
            axes,
            source_idx=None,
            log=True,
            **kwargs
    ):
        """Plots the spatial PDF ratio. If the signal PDF depends on the source,
        source_idx specifies the index of the source for which the PDF should
        get plotted.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the source
            hypotheses.
        axes : mpl.axes.Axes
            The matplotlib Axes object on which the PDF ratio should get drawn
            to.
        source_idx : int | None
            The index of the source for which the PDF ratio should get plotted.
            If set to None and the signal PDF depends on the source, index 0
            will be used.

        Additional Keyword Arguments
        ----------------------------
        Any additional keyword arguments will be passed to the `mpl.imshow`
        function.

        Returns
        -------
        img : instance of mpl.AxesImage
            The AxesImage instance showing the PDF ratio image.
        """
        if not isinstance(axes, Axes):
            raise TypeError(
                'The axes argument must be an instance of '
                'matplotlib.axes.Axes!')

        if source_idx is None:
            source_idx = 0

        # Define the binning for ra, dec, and sin_dec.
        delta_ra_deg = 0.5
        delta_dec_deg = 0.5

        raaxis = self._pdfratio.signalpdf.axes['ra']
        decaxis = self._pdfratio.signalpdf.axes['dec']

        # Create a grid of ratio in right-ascention and declination and fill it
        # with PDF ratio values from events that fall into these bins.
        # Use a binning for 1/2 degree.
        rabins = int(np.ceil(raaxis.length / np.deg2rad(delta_ra_deg)))
        decbins = int(np.ceil(decaxis.length / np.deg2rad(delta_dec_deg)))

        ratios = np.zeros((rabins, decbins), dtype=np.float64)

        ra_binedges = np.linspace(raaxis.vmin, raaxis.vmax, rabins+1)
        ra_bincenters = 0.5*(ra_binedges[:-1] + ra_binedges[1:])

        dec_binedges = np.linspace(decaxis.vmin, decaxis.vmax, decbins+1)
        dec_bincenters = 0.5*(dec_binedges[:-1] + dec_binedges[1:])

        # Generate events that fall into the ratio bins.
        events = DataFieldRecordArray(
            np.zeros(
                (ratios.size,),
                dtype=[('ira', np.int64),
                       ('ra', np.float64),
                       ('idec', np.int64),
                       ('dec', np.float64),
                       ('sin_dec', np.float64),
                       ('ang_err', np.float64)]))
        for (i, ((ira, ra), (idec, dec))) in enumerate(itertools.product(
                                                enumerate(ra_bincenters),
                                                enumerate(dec_bincenters))):
            events['ira'][i] = ira
            events['ra'][i] = ra
            events['idec'][i] = idec
            events['dec'][i] = dec
            events['sin_dec'][i] = np.sin(dec)
            events['ang_err'][i] = np.deg2rad(0.5)

        self._tdm.initialize_for_new_trial(src_hypo_group_manager, events)

        event_ratios = self._pdfratio.get_ratio(self._tdm)

        # Select only the ratios for the requested source.
        if event_ratios.ndim == 2:
            event_ratios = event_ratios[source_idx]

        ratios[events['ira'], events['idec']] = event_ratios

        (left, right, bottom, top) = (raaxis.vmin, raaxis.vmax,
                                      decaxis.vmin, decaxis.vmax)
        norm = LogNorm() if log else None
        img = axes.imshow(
            ratios.T,
            extent=(left, right, bottom, top),
            origin='lower',
            norm=norm,
            interpolation='none', **kwargs)
        axes.set_xlabel(raaxis.name)
        axes.set_ylabel(decaxis.name)
        axes.set_title(classname(self._pdfratio))

        return img



# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.py import (
    classname,
)


def get_bincenters_from_binedges(edges):
    """Calculates the bin center values from the given bin edge values.

    Parameters
    ----------
    edges : 1D numpy ndarray
        The (n+1,)-shaped 1D ndarray holding the bin edge values.

    Returns
    -------
    bincenters : 1D numpy ndarray
        The (n,)-shaped 1D ndarray holding the bin center values.
    """
    return 0.5*(edges[:-1] + edges[1:])


def get_binedges_from_bincenters(centers):
    """Calculates the bin edges from the given bin center values. The bin center
    values must be evenly spaced.

    Parameters
    ----------
    centers : 1D numpy ndarray
        The (n,)-shaped 1D ndarray holding the bin center values.

    Returns
    -------
    edges : 1D numpy ndarray
        The (n+1,)-shaped 1D ndarray holding the bin edge values.
    """
    d = np.diff(centers)
    if not np.all(np.isclose(np.diff(d), 0)):
        raise ValueError('The bin center values are not evenly spaced!')
    d = d[0]

    edges = np.zeros((len(centers)+1,), dtype=np.double)
    edges[:-1] = centers - d/2
    edges[-1] = centers[-1] + d/2

    return edges


def get_bin_indices_from_lower_and_upper_binedges(le, ue, values):
    """Returns the bin indices for the given values which must fall into bins
    defined by the given lower and upper bin edges.

    Note: The upper edge is not included in the bin.

    Parameters
    ----------
    le : (m,)-shaped 1D numpy ndarray
        The lower bin edges.
    ue : (m,)-shaped 1D numpy ndarray
        The upper bin edges.
    values : (n,)-shaped 1D numpy ndarray
        The values for which to get the bin indices.

    Returns
    -------
    idxs : (n,)-shaped 1D numpy ndarray
        The bin indices of the given values.
    """
    if len(le) != len(ue):
        raise ValueError(
            'The lower {} and upper {} edge arrays must be of the same '
            'size!'.format(
                len(le), len(ue)))

    if np.any(values < le[0]):
        invalid_values = values[values < le[0]]
        raise ValueError(
            '{} values ({}) are smaller than the lowest bin edge ({})!'.format(
                len(invalid_values), str(invalid_values), le[0]))
    if np.any(values >= ue[-1]):
        invalid_values = values[values >= ue[-1]]
        raise ValueError(
            '{} values ({}) are larger or equal than the largest bin edge '
            '({})!'.format(
                len(invalid_values), str(invalid_values), ue[-1]))

    m = (
        (values[:, np.newaxis] >= le[np.newaxis, :]) &
        (values[:, np.newaxis] < ue[np.newaxis, :])
    )
    idxs = np.nonzero(m)[1]

    return idxs


class BinningDefinition(object):
    """The BinningDefinition class provides a structure to hold histogram
    binning definitions for an analysis.
    """
    def __init__(
            self,
            name,
            binedges):
        """Creates a new binning definition object.

        Parameters
        ----------
        name : str
            The name of the binning definition.
        binedges : sequence of float
            The sequence of the bin edges, which should be used for the binning.
        """
        self.name = name
        self.binedges = binedges

    def __str__(self):
        """Pretty string representation.
        """
        s = f'{classname(self)}: {self._name}\n'
        s += str(self._binedges)
        return s

    def __eq__(self, other):
        """Checks if object ``other`` is equal to this BinningDefinition object.
        """
        if not isinstance(other, BinningDefinition):
            raise TypeError(
                'The other object in the equal comparison must be an instance '
                'of BinningDefinition! '
                f'Its current type is {classname(other)}.')
        if self.name != other.name:
            return False
        if np.any(self.binedges != other.binedges):
            return False

        return True

    @property
    def name(self):
        """The name of the binning setting. This must be an unique name
        for all the different binning settings used within a season.
        """
        return self._name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError(
                'The name must be of type str! '
                f'Its current type is {classname(name)}.')
        self._name = name

    @property
    def binedges(self):
        """The numpy.ndarray holding the bin edges.
        """
        return self._binedges

    @binedges.setter
    def binedges(self, arr):
        arr = np.atleast_1d(arr)
        self._binedges = np.array(arr, dtype=np.float64)

    @property
    def nbins(self):
        """The number of bins, based on the number of bin edges (minus 1).
        """
        return self._binedges.size - 1

    @property
    def bincenters(self):
        """The center values of the bins.
        """
        return 0.5*(self._binedges[:-1] + self._binedges[1:])

    @property
    def binwidths(self):
        """(read-only) The widths of the bins.
        """
        return np.diff(self._binedges)

    @property
    def lower_edge(self):
        """The lowest bin edge of the binning.
        """
        return self._binedges[0]

    @property
    def upper_edge(self):
        """The upper most edge of the binning.
        """
        return self._binedges[-1]

    @property
    def range(self):
        """The tuple (lower_edge, upper_edge) of the binning.
        """
        return (self.lower_edge, self.upper_edge)

    def any_data_out_of_range(self, data):
        """Checks if any of the given data is outside the range of this binning
        definition.

        Parameters
        ----------
        data : instance of ndarray
            The 1D ndarray with the data values to check.

        Returns
        -------
        outofrange : bool
            True if any data value is outside the binning range.
            False otherwise.
        """
        outofrange = np.any((data < self.lower_edge) |
                            (data > self.upper_edge))
        return outofrange

    def get_binwidth_from_value(self, value):
        """Returns the width of the bin the given value falls into.
        """
        idx = np.digitize(value, self._binedges) - 1

        bin_width = self.binwidths[idx]

        return bin_width

    def get_out_of_range_data(self, data):
        """Returns the data values which are outside the range of this binning
        definition.

        Parameters
        ----------
        data : instance of numpy.ndarray
            The 1D ndarray with the data values to check.

        Returns
        -------
        oor_data : instance of numpy.ndarray
            The 1D ndarray with data outside the range of this binning
            definition.
        """
        oor_mask = (
            (data < self.lower_edge) |
            (data > self.upper_edge)
        )
        oor_data = data[oor_mask]

        return oor_data

    def get_subset(self, lower_edge, upper_edge):
        """Creates a new BinningDefinition instance which contains only a subset
        of the bins of this BinningDefinition instance. The range of the subset
        is given by a lower and upper edge value.

        Parameters
        ----------
        lower_edge : float
            The lower edge value of the subset.
        upper_edge : float
            The upper edge value of the subset.

        Returns
        -------
        binning : instance of BinningDefinition
            The new instance of BinningDefinition holding the binning subset.
        """

        idxs = np.indices((len(self._binedges),))[0]
        m = (self._binedges >= lower_edge) & (self._binedges <= upper_edge)

        idx_lower = np.min(idxs[m])
        # Include the lower edge of the bin the lower_edge value falls into.
        if self._binedges[idx_lower] > lower_edge:
            idx_lower -= 1

        idx_upper = np.max(idxs[m])
        # Include the upper edge of the bin the upper_edge value falls into.
        if self._binedges[idx_upper] < upper_edge:
            idx_upper += 1

        new_binedges = self._binedges[idx_lower:idx_upper+1]

        return BinningDefinition(self._name, new_binedges)


class UsesBinning(object):
    """This is a classifier class that can be used to define, that a class uses
    binning information for one or more dimensions.

    This class defines the property ``binnings``, which is a list of
    BinningDefinition objects.

    This class provides the method ``has_same_binning_as(obj)`` to determine if
    a given object (that also uses binning) has the same binning.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Define the list of binning definition objects and a name->list_index
        # mapping for faster access.
        self._binnings = []
        self._binning_name2idx = {}

    @property
    def binnings(self):
        """(read-only) The list of BinningDefinition objects, one for each
        dimension.
        """
        return self._binnings

    @property
    def binning_ndim(self):
        """(read-only) The number of dimensions that uses binning.
        """
        return len(self._binnings)

    def has_same_binning_as(self, obj):
        """Checks if this object has the same binning as the given object.

        Parameters
        ----------
        obj : instance of UsesBinning
            The object that should be checked for same binning.

        Returns
        -------
        check : bool
            True if ``obj`` uses the same binning, False otherwise.
        """
        if not isinstance(obj, UsesBinning):
            raise TypeError(
                'The obj argument must be an instance of UsesBinning! '
                f'Its current type is {classname(obj)}.')

        for (self_binning, obj_binning) in zip(self.binnings, obj.binnings):
            if self_binning != obj_binning:
                return False

        return True

    def add_binning(self, binning, name=None):
        """Adds the given binning definition to the list of binnings.

        Parameters
        ----------
        binning : instance of BinningDefinition
            The binning definition to add.
        name : str | (default) None
            The name of the binning. If not None and it's different to the
            name of the given binning definition, a copy of the
            BinningDefinition object is made and the new name is set.
        """
        if not isinstance(binning, BinningDefinition):
            raise TypeError(
                'The binning argument must be an instance of '
                'BinningDefinition! '
                f'Its current type is {classname(binning)}.')

        # Create a copy of the BinningDefinition object if the name differs.
        if name is not None:
            if not isinstance(name, str):
                raise TypeError(
                    'The name argument must be of type str! '
                    f'Its current type is {classname(name)}.')
            if name != binning.name:
                binning = BinningDefinition(name, binning.binedges)

        self._binnings.append(binning)
        self._binning_name2idx[binning.name] = len(self._binnings)-1

    def get_binning(self, name):
        """Retrieves the binning definition of the given name.

        Parameters
        ----------
        name : str | int
            The name of the binning definition. A string specifies the name and
            an integer the dimension index.

        Returns
        -------
        binning : instance of BinningDefinition
            The binning definition of the given name.
        """
        if isinstance(name, str):
            if name not in self._binning_name2idx:
                raise KeyError(
                    f'The binning definition "{name}" is not defined!')
            binning = self._binnings[self._binning_name2idx[name]]
        elif isinstance(name, int):
            binning = self._binnings[name]
        else:
            raise TypeError(
                'The name argument must be of type str or int! '
                f'Its current type is {classname(name)}.')

        return binning



# -*- coding: utf-8 -*-

import logging
import queue
import time

import multiprocessing as mp
import numpy as np

from logging.handlers import (
    QueueHandler,
)

from skyllh.core.config import (
    HasConfig,
)
from skyllh.core.progressbar import (
    ProgressBar,
)
from skyllh.core.py import (
    classname,
)
from skyllh.core.random import (
    RandomStateService,
)
from skyllh.core.timing import (
    TimeLord,
)


def get_ncpu(
        cfg,
        local_ncpu,
):
    """Determines the number of CPUs to use for functions that support
    multi-processing.

    Parameters
    ----------
    cfg : instance of Config
        The instance of Config holding the local configuration.
    local_ncpu : int | None
        The local setting of the number of CPUs to use.

    Returns
    -------
    ncpu : int
        The number of CPUs to use by functions that allow multi-processing.
        If ``local_ncpu`` is set to None, the global NCPU setting is returned.
        If the global NCPU setting is None as well, the default value 1 is
        returned.
    """
    ncpu = local_ncpu
    if ncpu is None:
        ncpu = cfg['multiproc']['ncpu']
    if ncpu is None:
        ncpu = 1

    if not isinstance(ncpu, int):
        raise TypeError(
            'The ncpu setting must be of type int!')

    if ncpu < 1:
        raise ValueError(
            'The ncpu setting must be >= 1!')

    return ncpu


def parallelize(  # noqa: C901
        func,
        args_list,
        ncpu,
        rss=None,
        tl=None,
        ppbar=None,
):
    """Parallelizes the execution of the given function for different arguments.

    Parameters
    ----------
    func : callable
        The function which should be called with different arguments, which are
        given through the args_list argument. If the `rss` argument is not None,
        `func` requires an argument named `rss`.
    args_list : list of 2-element tuple
        The list of the different arguments for function ``func``. Each element
        of that list must be a 2-element tuple, where the first element is a
        tuple of the arguments of ``func``, and the second element is a
        dictionary with the keyword arguments of ``func``. If the `rss` argument
        is not None, `func` argument `rss` has to be omitted.
    ncpu : int
        The number of CPUs to use, i.e. the number of subprocesses to spawn.
    rss : RandomStateService | None
        The RandomStateService instance to use for generating random numbers.
    tl : instance of TimeLord | None
        The instance of TimeLord that should be used to time individual tasks.
    ppbar : instance of ProgressBar | None
        The possible parent ProgressBar instance.

    Returns
    -------
    result_list : list
        The list of the result values of ``func``, where each element of that
        list corresponds to the arguments element in ``args_list``.
    """
    # Define a wrapper function for the multiprocessing module that evaluates
    # ``func`` for a subset of `args_list` on a worker process.
    def worker_wrapper(
            func,
            sub_args_list,
            pid,
            rqueue,
            lqueue,
            squeue=None,
            rss=None,
            tl=None,
    ):
        """Wrapper function for the multiprocessing module that evaluates
        ``func`` for the subset ``sub_args_list`` of ``args_list`` on a worker
        process.

        Parameters
        ----------
        func : callable
            The function which should be called with different arguments, which
            are given through the sub_args_list argument. If the `rss` argument
            is not None `func` requires an argument named `rss`.
        sub_args_list : list of 2-element tuple
            The list of the different arguments for function ``func``. Each
            element of that list must be a 2-element tuple, where the first
            element is a tuple of the arguments of ``func``, and the second
            element is a dictionary with the keyword arguments of ``func``.
            If the `rss` argument is not None, `func` argument `rss` has to be
            omitted.
        pid : int
            The process ID that identifies the process in order to sort the
            results to the initial order of the function arguments.
        rqueue : multiprocessing.Queue
            The Queue instance where to put the function result in.
            If set to None, the result list will be returned.
        lqueue : multiprocessing.Queue
            The queue to hold generated log records by a given function.
        squeue : multiprocessing.Queue | None
            The Queue instance where to put in status information about finished
            tasks. Can be None to skip sending status information.
        rss : RandomStateService | None
            The RandomStateService instance to use for generating random
            numbers.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time individual
            tasks.
        """
        # Get the `QueueHandler` and update its log records queue.
        logger = logging.getLogger('skyllh')
        queue_handler = list(logger.handlers)[0]
        queue_handler.queue = lqueue

        result_list = []
        for (task_idx, (args, kwargs)) in enumerate(sub_args_list):
            if rss is not None:
                kwargs['rss'] = rss
            if tl is not None:
                kwargs['tl'] = tl
            result_list.append(func(*args, **kwargs))

            if squeue is not None:
                squeue.put((pid, task_idx))

        rqueue.put((pid, result_list, tl))

        # Put None object as the last log records queue item.
        lqueue.put_nowait(None)

    # Define a wrapper function that evaluates ``func`` for a subset of
    # `args_list` on the master process.
    def master_wrapper(
            pbar,
            sarr,
            func,
            sub_args_list,
            squeue=None,
            rss=None,
            tl=None,
    ):
        """This is the wrapper function for the master process.

        Parameters
        ----------
        pbar : instance of ProgressBar | None
            The instance of ProgressBar that should be used to display the
            progress if the current session is interactive.
        sarr : numpy record ndarray
            The status numpy record ndarray for all the processes. The length of
            that array must equal the number of processes, including the master
            process. Hence, the array index is the process id. The array must
            contain the following fields:
                n_finished_tasks : int
                    The number of finished tasks.
        func : callable
            The function which should be called with different arguments, which
            are given through the sub_args_list argument. If the `rss` argument
            is not None `func` requires an argument named `rss`.
        sub_args_list : list of 2-element tuple
            The list of the different arguments for function ``func``. Each
            element of that list must be a 2-element tuple, where the first
            element is a tuple of the arguments of ``func``, and the second
            element is a dictionary with the keyword arguments of ``func``.
            If the `rss` argument is not None, `func` argument `rss` has to be
            omitted.
        squeue : multiprocessing.Queue | None
            The status queue for the worker processes that should be used to
            receive status information about finished tasks.
        rss : RandomStateService | None
            The RandomStateService instance to use for generating random
            numbers.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time individual
            tasks.
        """
        result_list = []
        for (master_task_idx, (args, kwargs)) in enumerate(sub_args_list):
            if rss is not None:
                kwargs['rss'] = rss
            if tl is not None:
                kwargs['tl'] = tl
            result_list.append(func(*args, **kwargs))

            # Skip the rest, if we are not in an interactive session, hence
            # there is not progress bar.
            if not pbar.is_shown:
                continue

            sarr[0]['n_finished_tasks'] = master_task_idx + 1

            # Get possible status information from the worker processes.
            if squeue is not None:
                while not squeue.empty():
                    (pid, worker_task_idx) = squeue.get()
                    sarr[pid]['n_finished_tasks'] = worker_task_idx + 1

            # Calculate the total number of finished tasks.
            n_finished_tasks = np.sum(sarr['n_finished_tasks'])
            pbar.update(n_finished_tasks)

        return result_list

    # Create the progress bar if we are in an interactive session.
    pbar = ProgressBar(maxval=len(args_list), parent=ppbar).start()

    # Return result list if only one CPU is used.
    if ncpu == 1:
        sarr = np.zeros((1,), dtype=[('n_finished_tasks', np.int64)])
        result_list = master_wrapper(
            pbar, sarr, func, args_list, squeue=None, rss=rss, tl=tl)

        pbar.finish()

        return result_list

    # Multiple CPUs are used. Split the work across multiple processes.
    # We will use our own process (pid = 0) as a worker too.
    rqueue = mp.Queue()
    squeue = None
    if pbar.is_shown:
        squeue = mp.Queue()

    sub_args_list_list = np.array_split(np.array(args_list, dtype=object), ncpu)

    # Create a multiprocessing queue for each worker process.
    # Prepend it with None to be able to use `pid` as the list index.
    lqueue_list = [None] + [mp.Queue() for i in range(ncpu-1)]

    # Create a list of RandomStateService for each process if rss argument is
    # set.
    rss_list = [rss]
    if rss is None:
        rss_list += [None]*(ncpu-1)
    else:
        if not isinstance(rss, RandomStateService):
            raise TypeError(
                'The rss argument must be an instance of RandomStateService!')
        rss_list.extend([
            RandomStateService(seed=rss.random.randint(0, 2**32))
            for i in range(1, ncpu)
        ])

    # Create a list of TimeLord instances, one for each process if tl argument
    # is set.
    tl_list = [tl]
    if tl is None:
        tl_list += [None]*(ncpu-1)
    else:
        if not isinstance(tl, TimeLord):
            raise TypeError(
                'The tl argument must be an instance of TimeLord!')
        tl_list.extend([
            TimeLord()
            for i in range(1, ncpu)
        ])

    # Replace all existing main process handlers with the `QueueHandler`.
    # This allows storing all the log record generated by worker processes at
    # separate `multiprocessing.Queue` instances. After creating
    # worker processes revert handlers to the initial state.
    logger = logging.getLogger('skyllh')
    orig_handlers = list(logger.handlers)
    for orig_handler in orig_handlers:
        logger.removeHandler(orig_handler)
    queue_handler = QueueHandler(lqueue_list[0])
    logger.addHandler(queue_handler)

    processes = [mp.Process(
        target=worker_wrapper,
        args=(func, sub_args_list, pid, rqueue, lqueue_list[pid]),
        kwargs={'squeue': squeue, 'rss': rss_list[pid], 'tl': tl_list[pid]})
        for (pid, sub_args_list) in enumerate(sub_args_list_list) if pid > 0]

    # Start the processes.
    for proc in processes:
        proc.start()

    # Revert main process handlers to the initial state.
    logger.removeHandler(queue_handler)
    for orig_handler in orig_handlers:
        logger.addHandler(orig_handler)

    # Compute the first chunk in the main process.
    sarr = np.zeros((len(processes)+1,), dtype=[('n_finished_tasks', np.int64)])
    result_list_0 = master_wrapper(
        pbar, sarr, func, sub_args_list_list[0], squeue=squeue, rss=rss_list[0],
        tl=tl_list[0])

    # Initialize logger.
    logger = logging.getLogger(__name__)

    # Gather len(processes) results from the rqueue and join the process's
    # TimeLord instance with the main TimeLord instance.
    # Handle log records created by each process.
    pid_result_list_map = {0: result_list_0}
    for proc in processes:
        # Get the result record from the result queue.
        result_received = False
        proc_died = False
        while (result_received is False) and (proc_died is False):
            try:
                (pid, result_list, proc_tl) = rqueue.get(block=False)
                result_received = True
            except queue.Empty:
                # If this exception is raised, either the child process isn't
                # finished yet, or it dies due to an exception.
                if proc.exitcode is None:
                    # Child process hasn't finish yet.
                    # We'll wait a short moment.
                    time.sleep(0.01)
                elif proc.exitcode != 0:
                    proc_died = True
        if proc_died:
            raise RuntimeError(
                f'Child process {proc.pid} did not return with 0! '
                f'Exit code was {proc.exitcode}.')

        pid_result_list_map[pid] = result_list
        if tl is not None:
            tl.join(proc_tl)
        logger.debug(
            f'Beginning of worker process (pid={pid}) log records.')
        lqueue_end = False
        while not lqueue_end:
            record = lqueue_list[pid].get()
            if record is None:
                lqueue_end = True
            else:
                lqueue_logger = logging.getLogger(record.name)
                lqueue_logger.handle(record)
        logger.debug('Ending of worker process (pid=%d) log records.', pid)

    # Join all the processes.
    for proc in processes:
        proc.join()

    # Order the result lists.
    result_list = []
    for pid in range(len(pid_result_list_map)):
        result_list += pid_result_list_map[pid]

    pbar.finish()

    return result_list


class IsParallelizable(
        object,
):
    """Classifier class defining the ncpu property. Classes that derive from
    this class indicate, that they can make use of multi-processing on several
    CPUs at the same time.
    """

    def __init__(
            self,
            *args,
            ncpu=None,
            **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if not isinstance(self, HasConfig):
            raise TypeError(
                f'The class "{classname(self)}" is not derived from '
                'skyllh.core.config.HasConfig!')

        self.ncpu = ncpu

    @property
    def ncpu(self):
        """The number (int) of CPUs to utilize. It calls the ``get_ncpu``
        utility function with this property as argument. Hence, if this property
        is set to None, the global NCPU setting will take precedence.
        """
        return get_ncpu(self._cfg, self._ncpu)

    @ncpu.setter
    def ncpu(self, n):
        if n is not None:
            if not isinstance(n, int):
                raise TypeError(
                    'The ncpu property must be of type int!')
            if n < 1:
                raise ValueError(
                    'The ncpu property must be >= 1!')
        self._ncpu = n



# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.binning import (
    BinningDefinition,
)
from skyllh.core.flux_model import (
    FluxModel,
)
from skyllh.core.multiproc import (
    IsParallelizable,
    parallelize,
)
from skyllh.core.parameters import (
    ParameterGrid,
    ParameterGridSet,
)
from skyllh.core.pdf import (
    PDF,
    PDFSet,
    IsSignalPDF,
)
from skyllh.core.py import (
    classname,
)
from skyllh.core.smoothing import (
    SmoothingFilter,
)
from skyllh.i3.pdf import (
    I3EnergyPDF,
)


class SignalI3EnergyPDFSet(
        PDFSet,
        IsSignalPDF,
        PDF,
        IsParallelizable,
):
    """This is the signal energy PDF for IceCube. It creates a set of
    I3EnergyPDF objects for a discrete set of energy signal parameters. Energy
    signal parameters influence the source's flux model.
    """
    def __init__(
            self,
            cfg,
            data_mc,
            log10_energy_binning,
            sin_dec_binning,
            fluxmodel,
            param_grid_set,
            smoothing_filter=None,
            ncpu=None,
            ppbar=None,
            **kwargs,
    ):
        """Creates a new IceCube energy signal PDF for a given flux model and
        a set of parameter grids for the flux model.
        It creates a set of I3EnergyPDF objects for each signal parameter value
        permutation and stores it in an internal dictionary, where the hash of
        the parameters dictionary is the key.

        Parameters
        ----------
        cfg : instance of Config
            The instance of Config holding the local configuration.
        data_mc : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the monte-carlo data.
            The following data fields must exist:

            true_energy : float
                The true energy value of the data event.
            log_energy : float
                The base10-logarithm of the reconstructed energy value of the
                data event.
            sin_dec : float
                The declination of the data event.
            mcweight : float
                The monte-carlo weight value of the data events in unit
                GeV cm^2 sr.

        log10_energy_binning : instance of BinningDefinition
            The binning definition for the reconstructed energy binning in
            log10(E).
        sin_dec_binning : instance of BinningDefinition
            The binning definition for the binning in sin(declination).
        fluxmodel : instance of FluxModel
            The flux model to use to create the signal energy PDF.
        param_grid_set : instance of ParameterGridSet |
                         instance of ParameterGrid
            The set of parameter grids. A ParameterGrid instance for each
            energy parameter, for which an I3EnergyPDF object needs to be
            created.
        smoothing_filter : instance of SmoothingFilter | None
            The smoothing filter to use for smoothing the energy histogram.
            If ``None``, no smoothing will be applied.
        ncpu : int | None
            The number of CPUs to use to create the different I3EnergyPDF
            instances for the different parameter grid values.
            If set to ``None``, the configured default number of CPUs will be
            used.
        ppbar : instance of ProgressBar | None
            The instance of ProgressBar of the optional parent progress bar.
        """
        if isinstance(param_grid_set, ParameterGrid):
            param_grid_set = ParameterGridSet([param_grid_set])
        if not isinstance(param_grid_set, ParameterGridSet):
            raise TypeError(
                'The param_grid_set argument must be an instance of '
                'ParameterGrid or ParameterGridSet! But its type is '
                f'{classname(param_grid_set)}!')

        # We need to extend the parameter grids on the lower and upper end
        # by one bin to allow for the calculation of the interpolation. But we
        # will do this on a copy of the object.
        param_grid_set = param_grid_set.copy()
        param_grid_set.add_extra_lower_and_upper_bin()

        super().__init__(
            cfg=cfg,
            param_grid_set=param_grid_set,
            ncpu=ncpu,
            **kwargs)

        if not isinstance(log10_energy_binning, BinningDefinition):
            raise TypeError(
                'The log10_energy_binning argument must be an instance of '
                'BinningDefinition! '
                f'Its type is {classname(log10_energy_binning)}!')
        if not isinstance(sin_dec_binning, BinningDefinition):
            raise TypeError(
                'The sin_dec_binning argument must be an instance '
                'of BinningDefinition! '
                f'Its type is {classname(sin_dec_binning)}!')
        if not isinstance(fluxmodel, FluxModel):
            raise TypeError(
                'The fluxmodel argument must be an instance of FluxModel! '
                f'Its type is {classname(fluxmodel)}!')
        if (smoothing_filter is not None) and\
           (not isinstance(smoothing_filter, SmoothingFilter)):
            raise TypeError(
                'The smoothing_filter argument must be None or '
                'an instance of SmoothingFilter! '
                f'Its type is {classname(smoothing_filter)}!')

        # Create I3EnergyPDF objects for all permutations of the parameter
        # grid values.
        def create_I3EnergyPDF(
                cfg,
                data_log10_energy,
                data_sin_dec,
                data_mcweight,
                data_true_energy,
                log10_energy_binning,
                sin_dec_binning,
                smoothing_filter,
                fluxmodel,
                flux_unit_conv_factor,
                gridparams,
        ):
            """Creates an I3EnergyPDF object for the given flux model and flux
            parameters.

            Parameters
            ----------
            cfg : instance of Config
                The instance of Config holding the local configuration.
            data_log10_energy : 1d ndarray
                The base-10 logarithm of the reconstructed energy value of the
                data events.
            data_sin_dec : 1d ndarray
                The sin(dec) value of the the data events.
            data_mcweight : 1d ndarray
                The monte-carlo weight value of the data events.
            data_true_energy : 1d ndarray
                The true energy value of the data events.
            log10_energy_binning : instance of BinningDefinition
                The binning definition for the binning in log10(E).
            sin_dec_binning : instance of BinningDefinition
                The binning definition for the sin(declination).
            smoothing_filter : instance of SmoothingFilter | None
                The smoothing filter to use for smoothing the energy histogram.
                If ``None``, no smoothing will be applied.
            fluxmodel : instance of FluxModel
                The flux model to use to create the signal event weights.
            flux_unit_conv_factor : float
                The factor to convert the flux unit into the internal flux unit.
            gridparams : dict
                The dictionary holding the specific signal flux parameters.

            Returns
            -------
            i3energypdf : instance of I3EnergyPDF
                The created I3EnergyPDF instance for the given flux model and
                flux parameters.
            """
            # Create a copy of the FluxModel with the given flux parameters.
            # The copy is needed to not interfer with other CPU processes.
            myfluxmodel = fluxmodel.copy(newparams=gridparams)

            # Calculate the signal energy weight of the event. Note, that
            # because we create a normalized PDF, we can ignore all constants.
            # So we don't have to convert the flux unit into the internally used
            # flux unit.
            data_physicsweight = np.squeeze(myfluxmodel(E=data_true_energy))
            data_physicsweight *= flux_unit_conv_factor

            i3energypdf = I3EnergyPDF(
                cfg=cfg,
                pmm=None,
                data_log10_energy=data_log10_energy,
                data_sin_dec=data_sin_dec,
                data_mcweight=data_mcweight,
                data_physicsweight=data_physicsweight,
                log10_energy_binning=log10_energy_binning,
                sin_dec_binning=sin_dec_binning,
                smoothing_filter=smoothing_filter)

            return i3energypdf

        data_log10_energy = data_mc['log_energy']
        data_sin_dec = data_mc['sin_dec']
        data_mcweight = data_mc['mcweight']
        data_true_energy = data_mc['true_energy']

        flux_unit_conv_factor =\
            fluxmodel.to_internal_flux_unit()

        args_list = [
            (
                (cfg,
                 data_log10_energy,
                 data_sin_dec,
                 data_mcweight,
                 data_true_energy,
                 log10_energy_binning,
                 sin_dec_binning,
                 smoothing_filter,
                 fluxmodel,
                 flux_unit_conv_factor,
                 gridparams),
                {}
            )
            for gridparams in self.gridparams_list
        ]

        i3energypdf_list = parallelize(
            func=create_I3EnergyPDF,
            args_list=args_list,
            ncpu=self.ncpu,
            ppbar=ppbar)

        # Save all the I3EnergyPDF instances in the PDFSet registry with
        # the hash of the individual parameters as key.
        for (gridparams, i3energypdf) in zip(self.gridparams_list,
                                             i3energypdf_list):
            self.add_pdf(i3energypdf, gridparams)



# -*- coding: utf-8 -*-

import numpy as np

from scipy.interpolate import interp1d


def make_spline_1d(
        x,
        y,
        kind='linear',
        **kwargs):
    """Creates a 1D spline for the function y(x) using
    :class:`scipy.interpolate.interp1d`.

    Parameters
    ----------
    x : array_like
        The x values.
    y : array_like
        The y values.
    kind : str
        The kind of the spline. See the :class:`scipy.interpolate.interp1d`
        documentation for possible values. Default is ``'linear'``.
    **kwargs
        Additional keyword arguments are passed to the :class:`~scipy.interpolate.interp1d` function.

    Returns
    -------
    spline :
        The created 1D spline instance.
    """
    x = np.atleast_1d(x)
    y = np.atleast_1d(y)

    # The interp1d function requires unique x values. So we need to sort x in
    # increasing order and mask out repeating x values.
    xy = np.array(sorted(zip(x, y)), dtype=y.dtype)
    x = xy[:, 0]
    unique_x_mask = np.concatenate(([True], np.invert(
        x[1:] <= x[:-1])))
    x = x[unique_x_mask]
    y = xy[:, 1][unique_x_mask]

    spline = interp1d(
        x,
        y,
        kind=kind,
        copy=False,
        assume_sorted=True,
        **kwargs)

    return spline


class CatmullRomRegular1DSpline(
        object):
    """This class provides a one-dimensional Catmull-Rom spline which is a C^1
    continous spline, where the control points coincide with the data points.
    The x data points need to be equal distant.

    .. note::

        The first and last data point are not part of the splined curve!

    """

    def __init__(
            self,
            x,
            y,
            **kwargs,
    ):
        """Creates a new CatmullRom1DSpline instance.

        Parameters
        ----------
        x : instance of ndarray
            The x values of the data points.
        y : instance of ndarray
            The y values of the data points.
        """
        super().__init__(
            **kwargs)

        if len(x) != len(y):
            raise ValueError(
                f'The number of x ({len(x)}) and y ({len(y)}) data values '
                'need to be equal!')
        if len(x) < 4:
            raise ValueError(
                f'The number of data points ({len(x)}) must be at least 4!')

        unique_delta_x = np.unique(np.diff(x))
        if len(unique_delta_x) != 1:
            raise ValueError(
                'The data points must be equal distant in x!')

        self._delta_x = unique_delta_x[0]
        self._x_start = x[1]
        self._x_stop = x[-2]

        # Calculates the number of segments given the number of data points.
        # Since there are 4 points necessary per segment, we need to subtract 3.
        self._num_segments = len(x) - 3

        # Calculate the required data for each segment.
        self._segment_data = []
        for seg_idx in range(self._num_segments):
            sl = slice(seg_idx, seg_idx+4)
            (t1, t2, t3) = self._calc_segment_coefficients(Px=x[sl], Py=y[sl])
            self._segment_data.append(
                (t1, t2, t3, x[sl], y[sl]))

    def _eval_for_valid_x(
            self,
            x,
    ):
        """Evaluates the spline given valid x-values in data coordinates.

        Parameters
        ----------
        x : instance of ndarray
            The instance of ndarray holding the valid values for which the
            spline should get evaluated.

        Returns
        -------
        y : instance of ndarray
            The instance of ndarray with the spline values at the given x
            values.
        """
        y = np.zeros((len(x),), dtype=np.float64)

        # Determine on which spline segment the data value belongs to.
        seg_idxs = np.empty((len(x),), dtype=np.int64)
        np.floor(
            (x - self._x_start) / self._delta_x,
            out=seg_idxs,
            casting='unsafe')
        m = x == self._x_stop
        seg_idxs[m] = self._num_segments - 1

        # Loop over the unique segments.
        for seg_idx in np.unique(seg_idxs):
            # Create a mask of the values belonging to this segment.
            mask = seg_idxs == seg_idx
            n_points = np.count_nonzero(mask)

            (t1, t2, t3, Px, Py) = self._segment_data[seg_idx]

            # Note: This linear relation between x and t is the reason why the
            # data point x values must be equal distant.
            t = (x[mask] - Px[1]) / (Px[2] - Px[1]) * (t2 - t1) + t1
            t = t.reshape(n_points, 1)

            P0 = (Px[0], Py[0])
            P1 = (Px[1], Py[1])
            P2 = (Px[2], Py[2])
            P3 = (Px[3], Py[3])

            t1_m_t0 = t1
            t2_m_t0 = t2
            t2_m_t1 = t2 - t1
            t3_m_t1 = t3 - t1
            t3_m_t2 = t3 - t2

            t_m_t0 = t
            t_m_t1 = t - t1
            t2_m_t = t2 - t
            t3_m_t = t3 - t

            A1 = (t1 - t) / (t1_m_t0) * P0 + (t_m_t0) / (t1_m_t0) * P1
            A2 = (t2_m_t) / (t2_m_t1) * P1 + (t_m_t1) / (t2_m_t1) * P2
            A3 = (t3_m_t) / (t3_m_t2) * P2 + (t - t2) / (t3_m_t2) * P3
            B1 = (t2_m_t) / (t2_m_t0) * A1 + (t_m_t0) / (t2_m_t0) * A2
            B2 = (t3_m_t) / (t3_m_t1) * A2 + (t_m_t1) / (t3_m_t1) * A3
            seg_points = t2_m_t / t2_m_t1 * B1 + t_m_t1 / t2_m_t1 * B2

            y[mask] = seg_points[:, 1]

        return y

    def __call__(
            self,
            x,
            oor_value=np.nan,
    ):
        """Evaluates the spline given x-values in data coordinates.

        Parameters
        ----------
        x : instance of ndarray
            The instance of ndarray holding the values for which the spline
            should get evaluated.

        Returns
        -------
        y : instance of ndarray
            The instance of ndarray with the spline values at the given x
            values.
        """
        x = np.atleast_1d(x)

        m_valid_x = (x >= self._x_start) & (x <= self._x_stop)

        y = np.full((len(x),), oor_value, dtype=np.float64)
        y[m_valid_x] = self._eval_for_valid_x(x=x[m_valid_x])

        return y

    def _calc_tj(
            self,
            ti,
            Pi_x,
            Pi_y,
            Pj_x,
            Pj_y,
    ):
        """Calculates the next segment coefficient ``tj`` given the previous
        segment coefficient ``ti`` and the previous and next data point
        ``(Pi_x, Pi_y)`` and ``(Pj_x, Pj_y)``, respectively.

        Parameters
        ----------
        ti : float
            The previous segment coefficient.
        Pi_x : float
            The x-value of the previous data point.
        Pi_y : float
            The y-value of the previous data point.
        Pj_x : float
            The x-value of the next data point.
        Pj_y : float
            The y-value of the next data point.

        Returns
        -------
        tj : float
            The next segment coefficient.
        """
        dx = Pj_x - Pi_x
        dy = Pj_y - Pi_y
        tj = ti + np.sqrt(np.sqrt(dx*dx + dy*dy))

        return tj

    def _calc_segment_coefficients(
            self,
            Px,
            Py,
    ):
        """Calculates the segment coefficients t1, t2, and t3 given the 4
        data (control) points of the segment. The coefficient t0 is 0 by
        definition.

        Parameters
        ----------
        Px : instance of ndarray
            The (4,)-shaped numpy ndarray holding the 4 x-values of the
            segment's data points.
        Py : instance of ndarray
            The (4,)-shaped numpy ndarray holding the 4 y-values of the
            segment's data points.
        """
        t0 = 0
        t1 = self._calc_tj(
            ti=t0, Pi_x=Px[0], Pi_y=Py[0], Pj_x=Px[1], Pj_y=Py[1])
        t2 = self._calc_tj(
            ti=t1, Pi_x=Px[1], Pi_y=Py[1], Pj_x=Px[2], Pj_y=Py[2])
        t3 = self._calc_tj(
            ti=t2, Pi_x=Px[2], Pi_y=Py[2], Pj_x=Px[3], Pj_y=Py[3])

        return (t1, t2, t3)



"""
The minimizer module provides functionality for the minimization process of
a function.
"""
import abc
import logging
import scipy.optimize

import numpy as np

from skyllh.core.config import (
    HasConfig,
)
from skyllh.core.parameters import (
    ParameterSet,
)
from skyllh.core.py import (
    classname,
)


logger = logging.getLogger(__name__)


class MinimizerImpl(
        HasConfig,
        metaclass=abc.ABCMeta,
):
    """Abstract base class for a minimizer implementation. It defines the
    interface between the implementation and the Minimizer class.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abc.abstractmethod
    def minimize(
            self,
            initials,
            bounds,
            func,
            func_args=None,
            **kwargs,
    ):
        """This method is supposed to minimize the given function with the given
        initials.

        Parameters
        ----------
        initials : 1D (N_fitparams)-shaped numpy ndarray
            The ndarray holding the initial values of all the fit parameters.
        bounds : 2D (N_fitparams,2)-shaped numpy ndarray
            The ndarray holding the boundary values (vmin, vmax) of the fit
            parameters.
        func : callable
            The function that should get minimized.
            The call signature must be

                ``__call__(x, *args)``

            The return value of ``func`` is minimizer implementation dependent.
        func_args : sequence | None
            Optional sequence of arguments for ``func``.

        Additional Keyword Arguments
        ----------------------------
        Additional keyword arguments include options for this minimizer
        implementation. These are implementation dependent.

        Returns
        -------
        xmin : 1D ndarray
            The array containing the function parameter values at the function's
            minimum.
        fmin : float
            The function value at its minimum.
        status : dict
            The status dictionary with information about the minimization
            process.
        """
        pass

    @abc.abstractmethod
    def get_niter(self, status):
        """This method is supposed to return the number of iterations that were
        required to find the minimum.

        Parameters
        ----------
        status : dict
            The dictionary with the status information about the last
            minimization process.

        Returns
        -------
        niter : int
            The number of iterations needed to find the minimum.
        """
        pass

    @abc.abstractmethod
    def has_converged(self, status):
        """This method is supposed to analyze the status information dictionary
        if the last minimization process has converged.

        Parameters
        ----------
        status : dict
            The dictionary with the status information about the last
            minimization process.

        Returns
        -------
        converged : bool
            The flag if the minimization has converged (True), or not (False).
        """
        pass

    @abc.abstractmethod
    def is_repeatable(self, status):
        """This method is supposed to analyze the status information dictionary
        if the last minimization process can be repeated to obtain a better
        minimum.

        Parameters
        ----------
        status : dict
            The dictionary with the status information about the last
            minimization process.

        Returns
        -------
        repeatable : bool
            The flag if the minimization process can be repeated to obtain a
            better minimum.
        """
        pass


class ScipyMinimizerImpl(
        MinimizerImpl
):
    """Wrapper for `scipy.optimize.minimize`"""

    def __init__(
            self,
            method: str,
            **kwargs,
    ) -> None:
        """Creates a new instance of ScipyMinimizerImpl.

        Parameters
        ----------
        method : str
            The minimizer method to use. See the documentation for the method
            argument of the :func:`scipy.optimize.minimize` function for
            possible values.
        """
        super().__init__(
            **kwargs)

        self._method = method

    def minimize(
            self,
            initials,
            bounds,
            func,
            func_args=None,
            **kwargs,
    ):
        """Minimizes the given function ``func`` with the given initial function
        argument values ``initials``.

        Parameters
        ----------
        initials : 1D numpy ndarray
            The ndarray holding the initial values of all the fit parameters.
        bounds : 2D (N_fitparams,2)-shaped numpy ndarray
            The ndarray holding the boundary values (vmin, vmax) of the fit
            parameters.
        func : callable
            The function that should get minimized.
            The call signature must be

                ``__call__(x, *args)``

            The return value of ``func`` must be (f, grads), the function value
            at the function arguments ``x`` and the ndarray with the values of
            the function gradient for each fit parameter, if the
            ``func_provides_grads`` keyword argument option is set to True.
            If set to False, ``func`` must return only the function value.
        func_args : sequence | None
            Optional sequence of arguments for ``func``.

        Additional Keyword Arguments
        ----------------------------

        Additional keyword arguments include options for this minimizer
        implementation. Possible options are:

            func_provides_grads : bool
                Flag if the function ``func`` also returns its gradients.
                Default is ``True``.


        Any additional keyword arguments are passed on to the underlaying
        :func:`scipy.optimize.minimize` minimization function.

        Returns
        -------
        xmin : instance of numpy.ndarray
            The 1D array containing the function arguments at the function's
            minimum.
        fmin : float
            The function value at its minimum.
        res : instance of scipy.optimize.OptimizeResult
            The scipy OptimizeResult.
        """

        method_supports_bounds = False

        # constraints: List[Dict[str, Any]]
        constraints = None

        # Check if method allows for bounds
        if self._method in ["L-BFGS-B", "TNC", "SLSQP"]:
            method_supports_bounds = True
        elif self._method == "COBYLA":
            # COBYLA doesn't allow for bounds, but we can convert bounds
            # to a linear constraint

            constraints = []
            for (bound_num, bound) in enumerate(bounds):
                lower, upper = bound
                lc = {"type": "ineq",
                      "fun": lambda x, lb=lower, i=bound_num: x[i] - lb}
                uc = {"type": "ineq",
                      "fun": lambda x, ub=upper, i=bound_num: ub - x[i]}
                constraints.append(lc)
                constraints.append(uc)
            bounds = None

        if (bounds is not None) and (not method_supports_bounds):
            logger.warn(
                f'Selected minimization method "{self._method}" does not '
                'support bounds. Continue at your own risk!')
            bounds = None

        if func_args is None:
            func_args = tuple()
        if kwargs is None:
            kwargs = {}

        func_provides_grads = kwargs.pop('func_provides_grads', True)

        res = scipy.optimize.minimize(
            func,
            initials,
            bounds=bounds,
            constraints=constraints,
            args=func_args,
            jac=func_provides_grads,
            **kwargs)

        return (res.x, res.fun, res)

    def get_niter(self, status):
        """Returns the number of iterations needed to find the minimum.

        Parameters
        ----------
        status : dict
            The dictionary with the status information about the minimization
            process.

        Returns
        -------
        niter : int
            The number of iterations needed to find the minimum.
        """
        return status['nit']

    def has_converged(self, status):
        """Analyzes the status information dictionary if the minimization
        process has converged. By definition the minimization process has
        converged if ``status['warnflag']`` equals 0.

        Parameters
        ----------
        status : dict
            The dictionary with the status information about the minimization
            process.

        Returns
        -------
        converged : bool
            The flag if the minimization has converged (True), or not (False).
        """
        return bool(status["success"])

    def is_repeatable(self, status):
        """Checks if the minimization process can be repeated to get a better
        result.

        TODO: Method specific checks. For now just return False


        Parameters
        ----------
        status : dict
            The dictionary with the status information about the last
            minimization process.

        Returns
        -------
        repeatable : bool
            The flag if the minimization process can be repeated to obtain a
            better minimum.
        """
        return False


class LBFGSMinimizerImpl(
        MinimizerImpl
):
    """The LBFGSMinimizerImpl class provides the minimizer implementation for
    L-BFG-S minimizer used from the :mod:`scipy.optimize` module.
    """

    def __init__(
            self,
            ftol=1e-6,
            pgtol=1e-5,
            maxls=100,
            iprint=-1,
            **kwargs,
    ):
        """Creates a new L-BGF-S minimizer instance to minimize the given
        likelihood function with its given partial derivatives.

        Parameters
        ----------
        ftol : float
            The function value tolerance.
        pgtol : float
            The gradient value tolerance.
        maxls : int
            The maximum number of line search steps for an iteration.
        iprint : int
            Controls the frequency of output. See scipy documentation for more
            information.
        """
        super().__init__(**kwargs)

        self._ftol = ftol
        self._pgtol = pgtol
        self._maxls = maxls
        self._iprint = iprint

        self._fmin_l_bfgs_b = scipy.optimize.fmin_l_bfgs_b

    def minimize(
            self,
            initials,
            bounds,
            func,
            func_args=None,
            **kwargs,
    ):
        """Minimizes the given function ``func`` with the given initial function
        argument values ``initials``.

        Parameters
        ----------
        initials : 1D numpy ndarray
            The ndarray holding the initial values of all the fit parameters.
        bounds : 2D (N_fitparams,2)-shaped numpy ndarray
            The ndarray holding the boundary values (vmin, vmax) of the fit
            parameters.
        func : callable
            The function that should get minimized.
            The call signature must be

                ``__call__(x, *args)``

            The return value of ``func`` must be (f, grads), the function value
            at the function arguments ``x`` and the ndarray with the values of
            the function gradient for each fit parameter, if the
            ``func_provides_grads`` keyword argument option is set to True.
            If set to False, ``func`` must return only the function value.
        func_args : sequence | None
            Optional sequence of arguments for ``func``.

        Additional Keyword Arguments
        ----------------------------
        Additional keyword arguments include options for this minimizer
        implementation. Possible options are:

            func_provides_grads : bool
                Flag if the function ``func`` also returns its gradients.
                Default is ``True``.


        Any additional keyword arguments are passed on to the underlying
        :func:`scipy.optimize.fmin_l_bfgs_b` minimization function.

        Returns
        -------
        xmin : 1D ndarray
            The array containing the function arguments at the function's
            minimum.
        fmin : float
            The function value at its minimum.
        status : dict
            The status dictionary with information about the minimization
            process. The following information are provided:

            niter : int
                The number of iterations needed to find the minimum.
            warnflag : int
                The warning flag indicating if the minimization did converge.
                The possible values are:

                    0: The minimization converged.
        """
        if func_args is None:
            func_args = tuple()
        if kwargs is None:
            kwargs = {}

        if 'factr' not in kwargs:
            kwargs['factr'] = self._ftol / np.finfo(float).eps
        if 'pgtol' not in kwargs:
            kwargs['pgtol'] = self._pgtol
        if 'maxls' not in kwargs:
            kwargs['maxls'] = self._maxls
        if 'iprint' not in kwargs:
            kwargs['iprint'] = self._iprint

        func_provides_grads = kwargs.pop('func_provides_grads', True)

        (xmin, fmin, status) = self._fmin_l_bfgs_b(
            func, initials,
            bounds=bounds,
            args=func_args,
            approx_grad=not func_provides_grads,
            **kwargs
        )

        return (xmin, fmin, status)

    def get_niter(
            self,
            status,
    ):
        """Returns the number of iterations needed to find the minimum.

        Parameters
        ----------
        status : dict
            The dictionary with the status information about the minimization
            process.

        Returns
        -------
        niter : int
            The number of iterations needed to find the minimum.
        """
        return status['nit']

    def has_converged(
            self,
            status,
    ):
        """Analyzes the status information dictionary if the minimization
        process has converged. By definition the minimization process has
        converged if ``status['warnflag']`` equals 0.

        Parameters
        ----------
        status : dict
            The dictionary with the status information about the minimization
            process.

        Returns
        -------
        converged : bool
            The flag if the minimization has converged (True), or not (False).
        """
        if status['warnflag'] == 0:
            return True
        return False

    def is_repeatable(
            self,
            status,
    ):
        """Checks if the minimization process can be repeated to get a better
        result. It's repeatable if

            `status['warnflag'] == 2 and 'FACTR' in str(status['task'])`

        Parameters
        ----------
        status : dict
            The dictionary with the status information about the last
            minimization process.

        Returns
        -------
        repeatable : bool
            The flag if the minimization process can be repeated to obtain a
            better minimum.
        """
        if status['warnflag'] == 2:
            task = str(status['task'])
            if 'FACTR' in task:
                return True
            if 'ABNORMAL_TERMINATION_IN_LNSRCH' in task:
                # This is causes most probably by starting the minimization at
                # a parameter boundary.
                return True
        return False


class NR1dNsMinimizerImpl(
        MinimizerImpl
):
    """The NR1dNsMinimizerImpl class provides a minimizer implementation for the
    Newton-Raphson method for finding the minimum of a one-dimensional R1->R1
    function, i.e. a function that depends solely on one parameter, the number of
    signal events ns.
    """

    def __init__(
            self,
            ns_tol=1e-3,
            max_steps=100,
            **kwargs,
    ):
        """Creates a new NRNs minimizer instance to minimize the given
        likelihood function with its given partial derivatives.

        Parameters
        ----------
        ns_tol : float
            The tolerance / precision for the ns parameter value.
        max_steps : int
            The maximum number of NR steps. If max_step is reached,
            the fit is considered NOT converged.
        """
        super().__init__(**kwargs)

        self.ns_tol = ns_tol
        self.max_steps = max_steps

    def minimize(  # noqa: C901
            self,
            initials,
            bounds,
            func,
            func_args=None,
            **kwargs,
    ):
        """Minimizes the given function ``func`` with the given initial function
        argument values ``initials``. This minimizer implementation will only
        vary the first parameter. All other parameters will be set to their
        initial value.

        Parameters
        ----------
        initials : 1D numpy ndarray
            The ndarray holding the initial values of all the fit parameters.
        bounds : 2D (N_fitparams,2)-shaped numpy ndarray
            The ndarray holding the boundary values (vmin, vmax) of the fit
            parameters.
        func : callable
            The function that should get minimized.
            The call signature must be

                ``__call__(x, *args)``

            The return value of ``func`` must be (f, grad, grad2), i.e. the
            function value at the function arguments ``x``, the value of the
            function first derivative for the one fit parameter, and the value
            of the second derivative for the one fit parameter.
        func_args : sequence | None
            Optional sequence of arguments for ``func``.

        Additional Keyword Arguments
        ----------------------------
        There are no additional options defined for this minimization
        implementation.

        Returns
        -------
        xmin : 1D ndarray
            The array containing the function parameter values at the function's
            minimum.
        fmin : float
            The function value at its minimum.
        status : dict
            The status dictionary with information about the minimization
            process. The following information are provided:

            niter : int
                The number of iterations needed to find the minimum.
            last_nr_step : float
                The Newton-Raphson step size of the last iteration.
            warnflag : int
                The warning flag indicating if the minimization did converge.
                The possible values are:

                    -1: The function minimum is above the upper bound of the
                        parameter value. Convergence forced at upper bound.
                    -2: The function minimum is below the lower bound of the
                        parameter value. Convergence forced at lower bound.
                    0: The minimization converged with a iteration step size
                       smaller than the specified precision.
                    1: The minimization did NOT converge within self.max_steps
                       number of steps

            warnreason: str
                The description for the set warn flag.

        """
        if func_args is None:
            func_args = tuple()

        (ns_min, ns_max) = bounds[0]
        if ns_min > initials[0]:
            raise ValueError(
                f'The initial value for ns ({initials[0]:g}) must be equal or '
                f'greater than the minimum bound value for ns ({ns_min:g})')

        ns_tol = self.ns_tol

        niter = 0
        x = np.copy(initials).astype(np.float64)
        ns = x[0]

        # Initialize stepsize to be larger than ns tolerance.
        # Also initialize first derivative to large value.
        # Want to perform at least one NR iteration.
        step = ns_tol + 1
        fprime = 1000
        # NR does not guarantee convergence, thus limit iterations.
        max_steps = self.max_steps
        status = {
            'warnflag': 0,
            'warnreason': '',
        }
        f = None
        at_boundary = False

        # We do the minimization process while the precision of ns is not
        # reached yet or the function is still rising or falling fast, i.e. the
        # minimum is in a deep well.
        # In case the optimum is found outside the bounds on ns the best fit
        # will be set to the boundary value and the fit considered converged.
        while ((ns_tol < np.fabs(step)) or (np.fabs(fprime) > 1.e-1)) and\
              (niter < max_steps):

            x[0] = ns
            (f, fprime, fprimeprime) = func(x, *func_args)
            step = -fprime / fprimeprime

            # Exit optimization if ns is at boundary but next step would be outside.
            if (ns == ns_min and step < 0.0) or (ns == ns_max and step > 0.0):
                at_boundary = True

                if ns == ns_min:
                    status['warnflag'] = -2
                    status['warnreason'] = (
                        'Function minimum is below the minimum bound of the '
                        'parameter value. Convergence forced at boundary.')
                elif ns == ns_max:
                    status['warnflag'] = -1
                    status['warnreason'] = (
                        'Function minimum is above the maximum bound of the '
                        'parameter value. Convergence forced at boundary.')
                break

            # Always perform step in ns as it improves the solution.
            ns += step

            # Do not allow ns outside boundaries.
            if ns < ns_min:
                ns = ns_min
            elif ns > ns_max:
                ns = ns_max

            # Increase counter since a step was taken.
            niter += 1

        x[0] = ns
        # Once converged evaluate function at minimum value unless
        # Convergence was forced at boundary
        # in which case function value is already known.
        if not at_boundary:
            (f, fprime, fprimeprime) = func(x, *func_args)

        if niter == max_steps:
            status['warnflag'] = 1
            status['warnreason'] = (
                f'NR optimization did not converge within {niter} NR steps.')

        status['niter'] = niter
        status['last_nr_step'] = step
        return (x, f, status)

    def get_niter(
            self,
            status,
    ):
        """Returns the number of iterations needed to find the minimum.

        Parameters
        ----------
        status : dict
            The dictionary with the status information about the minimization
            process.

        Returns
        -------
        niter : int
            The number of iterations needed to find the minimum.
        """
        return status['niter']

    def has_converged(
            self,
            status,
    ):
        """Analyzes the status information dictionary if the minimization
        process has converged. By definition the minimization process has
        converged if ``status['warnflag']`` is smaller or equal to 0.

        Parameters
        ----------
        status : dict
            The dictionary with the status information about the minimization
            process.

        Returns
        -------
        converged : bool
            The flag if the minimization has converged (True), or not (False).
        """
        if status['warnflag'] <= 0:
            return True

        return False

    def is_repeatable(self, status):
        """Checks if the minimization process can be repeated to get a better
        result. By definition of this minimization method, this method will
        always return ``False``.
        """
        return False


class NRNsScan2dMinimizerImpl(
        NR1dNsMinimizerImpl
):
    """The NRNsScan2dMinimizerImpl class provides a minimizer implementation for
    the R2->R1 function where the first dimension is minimized using the
    Newton-Raphson minimization method and the second dimension is scanned.
    """

    def __init__(
            self,
            p2_scan_step,
            ns_tol=1e-3,
            **kwargs,
    ):
        """Creates a new minimizer implementation instance.

        Parameters
        ----------
        p2_scan_step : float
            The step size for the scan of the second parameter of the function
            to minimize.
        ns_tol : float
            The tolerance / precision for the ns parameter value.
        """
        super().__init__(
            ns_tol=ns_tol,
            **kwargs)

        self.p2_scan_step = p2_scan_step

    def minimize(
            self,
            initials,
            bounds,
            func,
            func_args=None,
            **kwargs,
    ):
        """Minimizes the given function ``func`` with the given initial function
        argument values ``initials``. This minimizer implementation will only
        vary the first two parameters. The first parameter is the number of
        signal events, ns, and it is minimized using the Newton-Rapson method.
        The second parameter is scanned through its boundaries in step sizes of
        ``p2_scan_step``.

        Parameters
        ----------
        initials : 1D numpy ndarray
            The ndarray holding the initial values of all the fit parameters.
        bounds : 2D (N_fitparams,2)-shaped numpy ndarray
            The ndarray holding the boundary values (vmin, vmax) of the fit
            parameters.
        func : callable
            The function that should get minimized.
            The call signature must be

                ``__call__(x, *args)``

            The return value of ``func`` must be (f, grad, grad2), i.e. the
            function value at the function arguments ``x``, the value of the
            function first derivative for the first fit parameter, and the value
            of the second derivative for the first fit parameter.
        func_args : sequence | None
            Optional sequence of arguments for ``func``.

        Additional Keyword Arguments
        ----------------------------
        There are no additional options defined for this minimization
        implementation.

        Returns
        -------
        xmin : 1D ndarray
            The array containing the function parameter values at the function's
            minimum.
        fmin : float
            The function value at its minimum.
        status : dict
            The status dictionary with information about the minimization
            process. The following information are provided:

            niter : int
                The number of iterations needed to find the minimum.
            last_nr_step : float
                The Newton-Raphson step size of the last iteration.
            p2_n_steps : int
                The number of scanning steps performed for the 2nd parameter.
            warnflag : int
                The warning flag indicating if the minimization did converge.
                The possible values are:

                    0: The minimization converged with a iteration step size
                       smaller than the specified precision.
                    1: The function minimum is below the minimum bound of the
                       parameter value. The last iteration's step size did not
                       achieve the specified precision.
                    2: The function minimum is above the maximum bound of the
                       parameter value. The last iteration's step size did not
                       achieve the specified precision.
            warnreason: str
                The description for the set warn flag.
        """
        p2_low = bounds[1][0]
        p2_high = bounds[1][1]
        p2_scan_values = np.linspace(
            p2_low, p2_high, int((p2_high-p2_low)/self.p2_scan_step)+1)

        logger.debug(
            'Minimize func by scanning 2nd parameter in '
            f'{len(p2_scan_values):d} steps with a step size of '
            f'{np.mean(np.diff(p2_scan_values)):g}')

        niter_total = 0
        best_xmin = None
        best_fmin = None
        best_status = None
        for p2_value in p2_scan_values:
            initials[1] = p2_value
            (xmin, fmin, status) = super().minimize(
                initials, bounds, func, func_args, **kwargs)
            niter_total += status['niter']
            if (best_fmin is None) or (fmin < best_fmin):
                best_xmin = xmin
                best_fmin = fmin
                best_status = status

        best_status['p2_n_steps'] = len(p2_scan_values)
        best_status['niter'] = niter_total

        return (best_xmin, best_fmin, best_status)


class Minimizer(
        object
):
    """The Minimizer class provides the general interface for minimizing a
    function. The class takes an instance of MinimizerImpl for a specific
    minimizer implementation.
    """

    def __init__(
            self,
            minimizer_impl,
            max_repetitions=100,
            **kwargs,
    ):
        """Creates a new Minimizer instance.

        Parameters
        ----------
        minimizer_impl : instance of MinimizerImpl
            The minimizer implementation for a specific minimizer algorithm.
        max_repetitions : int
            In case the minimization process did not converge at the first time
            this option specifies the maximum number of repetitions with
            different initials.
        """
        super().__init__(**kwargs)

        self.minimizer_impl = minimizer_impl
        self.max_repetitions = max_repetitions

    @property
    def minimizer_impl(self):
        """The instance of MinimizerImpl, which provides the implementation of
        the minimizer.
        """
        return self._minimizer_impl

    @minimizer_impl.setter
    def minimizer_impl(self, impl):
        if not isinstance(impl, MinimizerImpl):
            raise TypeError(
                'The minimizer_impl property must be an instance of '
                'MinimizerImpl!')
        self._minimizer_impl = impl

    @property
    def max_repetitions(self):
        """In case the minimization process did not converge at the first time
        this option specifies the maximum number of repetitions with
        different initials.
        """
        return self._max_repetitions

    @max_repetitions.setter
    def max_repetitions(self, n):
        if not isinstance(n, int):
            raise TypeError(
                'The maximal repetitions property must be of type int!')
        self._max_repetitions = n

    def minimize(
            self,
            rss,
            paramset,
            func,
            args=None,
            kwargs=None,
    ):
        """Minimizes the the given function ``func`` by calling the ``minimize``
        method of the minimizer implementation.

        After the minimization process it calls the ``has_converged`` and
        ``is_repeatable`` methods of the minimizer implementation to determine
        if an additional minimization attempt has to be performed.
        This is repeated until the minimization process did converge or if the
        maximal number of repetitions has occurred.

        Parameters
        ----------
        rss : RandomStateService instance
            The RandomStateService instance to draw random numbers from.
        paramset : instance of ParameterSet
            The ParameterSet instances holding the floating parameters of the
            function ``func``.
        func : callable ``f(x, *args)``
            The function to be minimized. It must have the call signature

                ``__call__(x, *args)``

            The return value of ``func`` is minimizer implementation dependent.
        args : sequence of arguments for ``func`` | None
            The optional sequence of arguments for ``func``.
        kwargs : dict | None
            The optional dictionary with keyword arguments for the minimizer
            implementation minimize method.

        Returns
        -------
        xmin : 1d numpy ndarray
            The array holding the parameter values for which the function has
            a minimum.
        fmin : float
            The function value at its minimum.
        status : dict
            The status dictionary with information about the minimization
            process.
        """
        if not isinstance(paramset, ParameterSet):
            raise TypeError(
                'The paramset argument must be an instance of ParameterSet!')

        if kwargs is None:
            kwargs = dict()

        bounds = paramset.floating_param_bounds
        initials = paramset.floating_param_initials
        logger.debug(f'Doing function minimization: initials: {initials}.')

        (xmin, fmin, status) = self._minimizer_impl.minimize(
            initials, bounds, func, args, **kwargs)

        reps = 0
        while (not self._minimizer_impl.has_converged(status)) and\
              (self._minimizer_impl.is_repeatable(status)) and\
              (reps < self._max_repetitions):
            # The minimizer did not converge at the first time, but it is
            # possible to repeat the minimization process with different
            # initials to obtain a better result.

            # Create a new set of random parameter initials based on the
            # parameter bounds.
            initials = paramset.generate_random_floating_param_initials(
                rss=rss)

            logger.debug(
                'Previous rep ({}) status={}, new initials={}'.format(
                    reps, str(status), str(initials)))

            # Repeat the minimization process.
            (xmin, fmin, status) = self._minimizer_impl.minimize(
                initials, bounds, func, args, **kwargs)

            reps += 1

        # Store the number of repetitions in the status dictionary.
        status['skyllh_minimizer_n_reps'] = reps

        if not self._minimizer_impl.has_converged(status):
            raise ValueError(
                f'The minimizer did not converge after {reps:d} repetitions! '
                'The maximum number of repetitions is '
                f'{self._max_repetitions:d}. The status dictionary is '
                f'"{str(status)}".')

        # Check if any fit value is outside its bounds due to rounding errors by
        # the minimizer. If so, set those fit values to their respective bound
        # value and re-evaluate the function with the corrected fit values.
        condmin = xmin < bounds[:, 0]
        condmax = xmin > bounds[:, 1]
        if np.any(condmin) or np.any(condmax):
            xmin = np.where(condmin, bounds[:, 0], xmin)
            xmin = np.where(condmax, bounds[:, 1], xmin)
            if args is None:
                args = tuple()
            (fmin, grads) = func(xmin, *args)

        logger.debug(
            '%s (%s): Minimized function: %d iterations, %d repetitions, '
            'xmin=%s' % (
                classname(self), classname(self._minimizer_impl),
                self._minimizer_impl.get_niter(status), reps, str(xmin)))

        return (xmin, fmin, status)



# -*- coding: utf-8 -*-

import numpy as np

from scipy import (
    integrate,
    interpolate,
)

from skyllh.core.binning import (
    get_bincenters_from_binedges,
)


class FctSpline1D(object):
    """Class to represent a 1D function spline using the PchipInterpolator
    class from scipy.

    The evaluate the spline, use the ``__call__`` method.
    """

    def __init__(
            self,
            f,
            x_binedges,
            norm=False,
            **kwargs):
        """Creates a new 1D function spline using the PchipInterpolator
        class from scipy.

        Parameters
        ----------
        f : (n_x,)-shaped 1D numpy ndarray
            The numpy ndarray holding the function values at the bin centers.
        x_binedges : (n_x+1,)-shaped 1D numpy ndarray
            The numpy ndarray holding the bin edges of the x-axis.
        norm : bool
            Whether to precalculate and save normalization internally.
        """
        super().__init__(**kwargs)

        self.x_binedges = np.copy(x_binedges)

        self.x_min = self.x_binedges[0]
        self.x_max = self.x_binedges[-1]

        x = get_bincenters_from_binedges(self.x_binedges)

        self.spl_f = interpolate.PchipInterpolator(
            x, f, extrapolate=False
        )

        self.norm = None
        if norm:
            self.norm = integrate.quad(
                self.__call__,
                self.x_min,
                self.x_max,
                limit=200,
                full_output=1
            )[0]

    def __call__(
            self,
            x,
            oor_value=0):
        """Evaluates the spline at the given x values. For x-values
        outside the spline's range, the oor_value is returned.

        Parameters
        ----------
        x : (n_x,)-shaped 1D numpy ndarray
            The numpy ndarray holding the x values at which the spline should
            get evaluated.
        oor_value : float
            The value for out-of-range (oor) coordinates.

        Returns
        -------
        f : (n_x,)-shaped 1D numpy ndarray
            The numpy ndarray holding the evaluated values of the spline.
        """
        f = self.spl_f(x)
        f = np.where(np.isnan(f), oor_value, f)

        return f

    def evaluate(
            self,
            *args,
            **kwargs):
        """Alias for the __call__ method.
        """
        return self(*args, **kwargs)


class FctSpline2D(object):
    """Class to represent a 2D function spline using the RectBivariateSpline
    class from scipy.

    The spline is constructed in the log10 space of the function value to
    ensure a smooth spline.

    The evaluate the spline, use the ``__call__`` method.
    """

    def __init__(
            self,
            f,
            x_binedges,
            y_binedges,
            **kwargs):
        """Creates a new 2D function spline using the RectBivariateSpline
        class from scipy.

        Parameters
        ----------
        f : (n_x, n_y)-shaped 2D numpy ndarray
            he numpy ndarray holding the function values at the bin centers.
        x_binedges : (n_x+1,)-shaped 1D numpy ndarray
            The numpy ndarray holding the bin edges of the x-axis.
        y_binedges : (n_y+1,)-shaped 1D numpy ndarray
            The numpy ndarray holding the bin edges of the y-axis.
        """
        super().__init__(**kwargs)

        self.x_binedges = np.copy(x_binedges)
        self.y_binedges = np.copy(y_binedges)

        self.x_min = self.x_binedges[0]
        self.x_max = self.x_binedges[-1]
        self.y_min = self.y_binedges[0]
        self.y_max = self.y_binedges[-1]

        x = get_bincenters_from_binedges(self.x_binedges)
        y = get_bincenters_from_binedges(self.y_binedges)

        # Note: For simplicity we approximate zero bins with 1000x smaller
        # values than the minimum value. To do this correctly, one should store
        # the zero bins and return zero when those bins are requested.
        z = np.empty(f.shape, dtype=np.double)
        m = f > 0
        z[m] = np.log10(f[m])
        z[np.invert(m)] = np.min(z[m]) - 3

        self.spl_log10_f = interpolate.RectBivariateSpline(
            x, y, z, kx=3, ky=3, s=0)

    def __call__(
            self,
            x,
            y,
            oor_value=0):
        """Evaluates the spline at the given coordinates. For coordinates
        outside the spline's range, the oor_value is returned.

        Parameters
        ----------
        x : (n_x,)-shaped 1D numpy ndarray
            The numpy ndarray holding the x values at which the spline should
            get evaluated.
        y : (n_y,)-shaped 1D numpy ndarray
            The numpy ndarray holding the y values at which the spline should
            get evaluated.
        oor_value : float
            The value for out-of-range (oor) coordinates.

        Returns
        -------
        f : (n_x, n_y)-shaped 2D numpy ndarray
            The numpy ndarray holding the evaluated values of the spline.
        """
        m_x_oor = (x < self.x_min) | (x > self.x_max)
        m_y_oor = (y < self.y_min) | (y > self.y_max)

        (m_xx_oor, m_yy_oor) = np.meshgrid(m_x_oor, m_y_oor, indexing='ij')
        m_xy_oor = m_xx_oor | m_yy_oor

        f = np.power(10, self.spl_log10_f(x, y))
        f[m_xy_oor] = oor_value

        return f


def clip_grl_start_times(grl_data):
    """Make sure that the start time of a run is not smaller than the stop time
    of the previous run.

    Parameters
    ----------
    grl_data : instance of numpy structured ndarray
        The numpy structured ndarray of length N_runs, with the following
        fields:

        start : float
            The start time of the run.
        stop : float
            The stop time of the run.
    """
    start = grl_data['start']
    stop = grl_data['stop']

    m = (start[1:] - stop[:-1]) < 0
    new_start = np.where(m, stop[:-1], start[1:])

    grl_data['start'][1:] = new_start


def psi_to_dec_and_ra(
        rss,
        src_dec,
        src_ra,
        psi):
    """Generates random declinations and right-ascension coordinates for the
    given source location and opening angle `psi`.

    Parameters
    ----------
    rss : instance of RandomStateService
        The instance of RandomStateService to use for drawing random numbers.
    src_dec : float
        The declination of the source in radians.
    src_ra : float
        The right-ascension of the source in radians.
    psi : 1d ndarray of float
        The opening-angle values in radians.

    Returns
    -------
    dec : 1d ndarray of float
        The declination values.
    ra : 1d ndarray of float
        The right-ascension values.
    """

    psi = np.atleast_1d(psi)

    # Transform everything in radians and convert the source declination
    # to source zenith angle
    a = psi
    b = np.pi/2 - src_dec
    c = src_ra
    # Random rotation angle for the 2D circle
    t = rss.random.uniform(0, 2*np.pi, size=len(psi))

    # Parametrize the circle
    x = (
        (np.sin(a)*np.cos(b)*np.cos(c)) * np.cos(t) +
        (np.sin(a)*np.sin(c)) * np.sin(t) -
        (np.cos(a)*np.sin(b)*np.cos(c))
    )
    y = (
        -(np.sin(a)*np.cos(b)*np.sin(c)) * np.cos(t) +
        (np.sin(a)*np.cos(c)) * np.sin(t) +
        (np.cos(a)*np.sin(b)*np.sin(c))
    )
    z = (
        (np.sin(a)*np.sin(b)) * np.cos(t) +
        (np.cos(a)*np.cos(b))
    )

    # Convert back to right-ascension and declination.
    # This is to distinguish between diametrically opposite directions.
    zen = np.arccos(z)
    azi = np.arctan2(y, x)

    dec = np.pi/2 - zen
    ra = np.pi - azi

    return (dec, ra)


def create_energy_cut_spline(
        ds,
        exp_data,
        spl_smooth):
    """Create the spline for the declination-dependent energy cut
    that the signal generator needs for injection in the southern sky
    Some special conditions are needed for IC79 and IC86_I, because
    their experimental dataset shows events that should probably have
    been cut by the IceCube selection.
    """
    data_exp = exp_data.copy(keep_fields=['sin_dec', 'log_energy'])
    if ds.name == 'IC79':
        m = np.invert(np.logical_and(
            data_exp['sin_dec'] < -0.75,
            data_exp['log_energy'] < 4.2))
        data_exp = data_exp[m]
    if ds.name == 'IC86_I':
        m = np.invert(np.logical_and(
            data_exp['sin_dec'] < -0.2,
            data_exp['log_energy'] < 2.5))
        data_exp = data_exp[m]

    sin_dec_binning = ds.get_binning_definition('sin_dec')
    sindec_edges = sin_dec_binning.binedges
    min_log_e = np.zeros(len(sindec_edges)-1, dtype=float)
    for i in range(len(sindec_edges)-1):
        mask = np.logical_and(
            data_exp['sin_dec'] >= sindec_edges[i],
            data_exp['sin_dec'] < sindec_edges[i+1])
        min_log_e[i] = np.min(data_exp['log_energy'][mask])
    del data_exp
    sindec_centers = 0.5 * (sindec_edges[1:]+sindec_edges[:-1])

    spline = interpolate.UnivariateSpline(
        sindec_centers, min_log_e, k=2, s=spl_smooth)

    return spline



# -*- coding: utf-8 -*-

import numpy as np

from scipy import (
    integrate,
    interpolate,
)

from skyllh.analyses.i3.publicdata_ps.utils import (
    FctSpline2D,
)
from skyllh.core.binning import (
    get_bin_indices_from_lower_and_upper_binedges,
    get_bincenters_from_binedges,
)
from skyllh.core.storage import (
    create_FileLoader,
)


def load_effective_area_array(pathfilenames):
    """Loads the (nbins_decnu, nbins_log10enu)-shaped 2D effective
    area array from the given data file.

    Parameters
    ----------
    pathfilename : str | list of str
        The file name of the data file.

    Returns
    -------
    aeff_decnu_log10enu : (nbins_decnu, nbins_log10enu)-shaped 2D ndarray
        The ndarray holding the effective area for each
        (dec_nu,log10(E_nu/GeV)) bin.
    decnu_binedges_lower : (nbins_decnu,)-shaped ndarray
        The ndarray holding the lower bin edges of the dec_nu axis.
    decnu_binedges_upper : (nbins_decnu,)-shaped ndarray
        The ndarray holding the upper bin edges of the dec_nu axis.
    log10_enu_binedges_lower : (nbins_log10enu,)-shaped ndarray
        The ndarray holding the lower bin edges of the log10(E_nu/GeV) axis.
    log10_enu_binedges_upper : (nbins_log10enu,)-shaped ndarray
        The ndarray holding the upper bin edges of the log10(E_nu/GeV) axis.
    """
    loader = create_FileLoader(pathfilenames=pathfilenames)
    data = loader.load_data()
    renaming_dict = {
        'log10(E_nu/GeV)_min': 'log10_enu_min',
        'log10(E_nu/GeV)_max': 'log10_enu_max',
        'Dec_nu_min[deg]':     'decnu_min',
        'Dec_nu_max[deg]':     'decnu_max',
        'A_Eff[cm^2]':         'a_eff'
    }
    data.rename_fields(renaming_dict, must_exist=True)

    # Convert the true neutrino declination from degrees to radians.
    data['decnu_min'] = np.deg2rad(data['decnu_min'])
    data['decnu_max'] = np.deg2rad(data['decnu_max'])

    # Determine the binning for energy and declination.
    log10_enu_binedges_lower = np.unique(data['log10_enu_min'])
    log10_enu_binedges_upper = np.unique(data['log10_enu_max'])
    decnu_binedges_lower = np.unique(data['decnu_min'])
    decnu_binedges_upper = np.unique(data['decnu_max'])

    if len(log10_enu_binedges_lower) != len(log10_enu_binedges_upper):
        raise ValueError(
            'Cannot extract the log10(E/GeV) binning of the '
            'effective area from data file "{}". The number of lower and upper '
            'bin edges is not equal!'.format(str(loader.pathfilename_list)))
    if len(decnu_binedges_lower) != len(decnu_binedges_upper):
        raise ValueError(
            'Cannot extract the dec_nu binning of the effective '
            'area from data file "{}". The number of lower and upper bin edges '
            'is not equal!'.format(str(loader.pathfilename_list)))

    nbins_log10_enu = len(log10_enu_binedges_lower)
    nbins_decnu = len(decnu_binedges_lower)

    # Construct the 2d array for the effective area.
    aeff_decnu_log10enu = np.zeros(
        (nbins_decnu, nbins_log10_enu), dtype=np.double)

    decnu_idx = np.digitize(
        0.5*(data['decnu_min'] +
             data['decnu_max']),
        decnu_binedges_lower) - 1
    log10enu_idx = np.digitize(
        0.5*(data['log10_enu_min'] +
             data['log10_enu_max']),
        log10_enu_binedges_lower) - 1

    aeff_decnu_log10enu[decnu_idx, log10enu_idx] = data['a_eff']

    return (
        aeff_decnu_log10enu,
        decnu_binedges_lower,
        decnu_binedges_upper,
        log10_enu_binedges_lower,
        log10_enu_binedges_upper
    )


class PDAeff(object):
    """This class provides a representation of the effective area provided by
    the public data.
    """
    def __init__(
            self, pathfilenames, src_dec=None,
            min_log10enu=None, max_log10enu=None,
            **kwargs):
        """Creates an effective area instance by loading the effective area
        data from the given file.

        Parameters
        ----------
        pathfilenames : str | list of str
            The path file names of the effective area data file(s) which should
            be used for this public data effective area instance.
        src_dec : float | None
            The source declination in radians for which detection probabilities
            should get pre-calculated using the ``get_detection_prob_for_decnu``
            method.
        min_log10enu : float | None
            The minimum log10(E_nu/GeV) value that should be used for
            calculating the detection probability.
            If None, the lowest available neutrino energy bin edge of the
            effective area is used.
        max_log10enu : float | None
            The maximum log10(E_nu/GeV) value that should be used for
            calculating the detection probability.
            If None, the highest available neutrino energy bin edge of the
            effective area is used.
        """
        super().__init__(**kwargs)

        (
            self._aeff_decnu_log10enu,
            self._decnu_binedges_lower,
            self._decnu_binedges_upper,
            self._log10_enu_binedges_lower,
            self._log10_enu_binedges_upper
        ) = load_effective_area_array(pathfilenames)

        # Note: self._aeff_decnu_log10enu is numpy 2D ndarray of shape
        # (nbins_decnu, nbins_log10enu).

        # Cut the energies where all effective areas are zero.
        m = np.sum(self._aeff_decnu_log10enu, axis=0) > 0
        self._aeff_decnu_log10enu = self._aeff_decnu_log10enu[:, m]
        self._log10_enu_binedges_lower = self._log10_enu_binedges_lower[m]
        self._log10_enu_binedges_upper = self._log10_enu_binedges_upper[m]

        self._decnu_binedges = np.concatenate(
            (self._decnu_binedges_lower,
             self._decnu_binedges_upper[-1:])
        )
        self._log10_enu_binedges = np.concatenate(
            (self._log10_enu_binedges_lower,
             self._log10_enu_binedges_upper[-1:])
        )

        # Pre-calculate detection probabilities for a certain neutrino
        # declination if requested.
        if src_dec is not None:
            if min_log10enu is None:
                min_log10enu = self._log10_enu_binedges_lower[0]
            else:
                min_log10enu = max(
                    self._log10_enu_binedges_lower[0],
                    min_log10enu)

            if max_log10enu is None:
                max_log10enu = self._log10_enu_binedges_upper[-1]
            else:
                max_log10enu = min(
                    self._log10_enu_binedges_upper[-1],
                    max_log10enu)

            m = (
                (self.log10_enu_bincenters >= min_log10enu) &
                (self.log10_enu_bincenters < max_log10enu)
            )
            low_bin_edges = self._log10_enu_binedges_lower[m]
            high_bin_edges = self._log10_enu_binedges_upper[m]

            # Get the detection probability P(E_nu | sin(dec)) per bin.
            self.det_prob = self.get_detection_prob_for_decnu(
                src_dec,
                10**low_bin_edges,
                10**high_bin_edges,
                10**low_bin_edges[0],
                10**high_bin_edges[-1]
            )

    @property
    def decnu_binedges(self):
        """(read-only) The bin edges of the neutrino declination axis in
        radians.
        """
        return self._decnu_binedges

    @property
    def sin_decnu_binedges(self):
        """(read-only) The sin of the bin edges of the neutrino declination
        in radians.
        """
        return np.sin(self._decnu_binedges)

    @property
    def decnu_bincenters(self):
        """(read-only) The bin center values of the neutrino declination axis in
        radians.
        """
        return get_bincenters_from_binedges(self._decnu_binedges)

    @property
    def n_decnu_bins(self):
        """(read-only) The number of bins of the neutrino declination axis.
        """
        return len(self._decnu_binedges) - 1

    @property
    def log10_enu_binedges(self):
        """(read-only) The bin edges of the log10(E_nu/GeV) neutrino energy
        axis.
        """
        return self._log10_enu_binedges

    @property
    def log10_enu_binedges_lower(self):
        """(read-only) The lower binedges of the log10(E_nu/GeV) neutrino energy
        axis.
        """
        return self._log10_enu_binedges_lower

    @property
    def log10_enu_binedges_upper(self):
        """(read-only) The upper binedges of the log10(E_nu/GeV) neutrino energy
        axis.
        """
        return self._log10_enu_binedges_upper

    @property
    def log10_enu_bincenters(self):
        """(read-only) The bin center values of the log10(E_nu/GeV) neutrino
        energy axis.
        """
        return get_bincenters_from_binedges(self._log10_enu_binedges)

    @property
    def n_log10_enu_bins(self):
        """(read-only) The number of bins of the log10 neutrino energy axis.
        """
        return len(self._log10_enu_binedges) - 1

    @property
    def aeff_decnu_log10enu(self):
        """(read-only) The effective area in cm^2 as (n_decnu,n_log10enu)-shaped
        2D numpy ndarray.
        """
        return self._aeff_decnu_log10enu

    def create_sin_decnu_log10_enu_spline(self):
        """DEPRECATED!
        Creates a FctSpline2D object representing a 2D spline of the
        effective area in sin(dec_nu)-log10(E_nu/GeV)-space.

        Returns
        -------
        spl : FctSpline2D instance
            The FctSpline2D instance representing a spline in the
            sin(dec_nu)-log10(E_nu/GeV)-space.
        """
        spl = FctSpline2D(
            self._aeff_decnu_log10enu,
            self.sin_decnu_binedges,
            self.log10_enu_binedges
        )
        return spl

    def get_aeff_for_decnu(self, decnu):
        """Retrieves the effective area as function of log10_enu.

        Parameters
        ----------
        decnu : float
            The true neutrino declination.

        Returns
        -------
        aeff : (n,)-shaped numpy ndarray
            The effective area in cm^2 for the given true neutrino declination
            as a function of log10 true neutrino energy.
        """
        decnu_idx = np.digitize(decnu, self._decnu_binedges) - 1

        aeff = self._aeff_decnu_log10enu[decnu_idx]

        return aeff

    def get_detection_prob_for_decnu(
            self, decnu, enu_min, enu_max, enu_range_min, enu_range_max):
        """Calculates the detection probability for given true neutrino energy
        ranges for a given neutrino declination.

        Parameters
        ----------
        decnu : float
            The neutrino declination in radians.
        enu_min : float | ndarray of float
            The minimum energy in GeV.
        enu_max : float | ndarray of float
            The maximum energy in GeV.
        enu_range_min : float
            The minimum energy in GeV of the entire energy range.
        enu_range_max : float
            The maximum energy in GeV of the entire energy range.

        Returns
        -------
        det_prob : ndarray of float
            The neutrino energy detection probabilities for the given true
            enegry ranges.
        """
        enu_binedges = np.power(10, self.log10_enu_binedges)

        # Get the bin indices for the lower and upper energy range values.
        (lidx,) = get_bin_indices_from_lower_and_upper_binedges(
            enu_binedges[:-1],
            enu_binedges[1:],
            np.array([enu_range_min])
        )
        if enu_range_max >= enu_binedges[-1]:
            uidx = len(enu_binedges)-1
        else:
            (uidx,) = get_bin_indices_from_lower_and_upper_binedges(
                enu_binedges[:-1],
                enu_binedges[1:],
                np.array([enu_range_max])
            )
            # Note: The get_bin_indices_from_lower_and_upper_binedges function
            #       is based on the lower edges. So by definition the upper bin
            #       index is one too large.
            # uidx -= 1

        aeff = self.get_aeff_for_decnu(decnu)
        aeff = aeff[lidx:uidx]
        enu_binedges = enu_binedges[lidx:uidx+1]

        dE = np.diff(enu_binedges)

        daeff_dE = aeff / dE

        # Create a spline representation that spans the entire enu range.
        x = np.empty((len(enu_binedges)+1,), dtype=np.double)
        x[0] = enu_binedges[0]
        x[1:-1] = get_bincenters_from_binedges(enu_binedges)
        x[-1] = enu_binedges[-1]

        y = np.empty((len(enu_binedges)+1,), dtype=np.double)
        y[0] = daeff_dE[0]
        y[1:-1] = daeff_dE
        y[-1] = daeff_dE[-1]

        spl = interpolate.splrep(
            x,
            y,
            k=1,
            s=0
        )

        def _eval_spl_func(x):
            return interpolate.splev(x, spl, der=0, ext=1)

        norm = integrate.quad(
            _eval_spl_func,
            enu_range_min,
            enu_range_max,
            limit=200,
            full_output=1
        )[0]

        enu_min = np.atleast_1d(enu_min)
        enu_max = np.atleast_1d(enu_max)

        det_prob = np.empty((len(enu_min),), dtype=np.double)
        for i in range(len(enu_min)):
            integral = integrate.quad(
                _eval_spl_func,
                enu_min[i],
                enu_max[i],
                limit=200,
                full_output=1
            )[0]

            det_prob[i] = integral / norm

        return det_prob



# -*- coding: utf-8 -*-

import numpy as np

from scipy.stats import (
    gaussian_kde,
)

from skyllh.core.binning import (
    BinningDefinition,
    UsesBinning,
)
from skyllh.core.pdf import (
    EnergyPDF,
    IsBackgroundPDF,
    PDFAxis,
)
from skyllh.core.py import (
    classname,
)
from skyllh.core.smoothing import (
    UNSMOOTH_AXIS,
    SmoothingFilter,
    HistSmoothingMethod,
    NoHistSmoothingMethod,
    NeighboringBinHistSmoothingMethod,
)
from skyllh.core.storage import (
    DataFieldRecordArray,
)
from skyllh.core.timing import (
    TaskTimer,
)


class PDBackgroundI3EnergyPDF(
        EnergyPDF,
        IsBackgroundPDF,
        UsesBinning):
    """This is the base class for an IceCube specific energy background PDF for
    the public data.

    IceCube energy PDFs depend solely on the energy and the zenith angle, and
    hence, on the declination of the event.

    The IceCube energy PDF is modeled as a 1d histogram in energy,
    but for different sin(declination) bins, hence, stored as a 2d histogram.
    """

    _KDE_BW_NORTH = 0.4
    _KDE_BW_SOUTH = 0.32

    def __init__(
            self,
            data_logE,
            data_sinDec,
            data_mcweight,
            data_physicsweight,
            logE_binning,
            sinDec_binning,
            smoothing_filter,
            kde_smoothing=False,
            **kwargs):
        """Creates a new IceCube energy PDF object for the public data.

        Parameters
        ----------
        data_logE : instance of ndarray
            The 1d ndarray holding the log10(E) values of the events.
        data_sinDec : instance of ndarray
            The 1d ndarray holding the sin(dec) values of the events.
        data_mcweight : instance of ndarray
            The 1d ndarray holding the monte-carlo weights of the events.
            The final data weight will be the product of data_mcweight and
            data_physicsweight.
        data_physicsweight : instance of ndarray
            The 1d ndarray holding the physics weights of the events.
            The final data weight will be the product of data_mcweight and
            data_physicsweight.
        logE_binning : instance of BinningDefinition
            The binning definition for the log10(E) axis.
        sinDec_binning : instance of BinningDefinition
            The binning definition for the sin(declination) axis.
        smoothing_filter : instance of SmoothingFilter | None
            The smoothing filter to use for smoothing the energy histogram.
            If None, no smoothing will be applied.
        kde_smoothing : bool
            Apply a kde smoothing to the energy pdf for each bin in sin(dec).
            This is useful for signal injections, because it ensures that the
            background is not zero when injecting high energy events.
            Default: False.
        """
        super().__init__(
            pmm=None,
            **kwargs)

        # Define the PDF axes.
        self.add_axis(
            PDFAxis(
                name='log_energy',
                vmin=logE_binning.lower_edge,
                vmax=logE_binning.upper_edge))
        self.add_axis(
            PDFAxis(
                name='sin_dec',
                vmin=sinDec_binning.lower_edge,
                vmax=sinDec_binning.upper_edge))

        self.add_binning(logE_binning, 'log_energy')
        self.add_binning(sinDec_binning, 'sin_dec')

        # Create the smoothing method instance tailored to the energy PDF.
        # We will smooth only the first axis (logE).
        if (smoothing_filter is not None) and\
           (not isinstance(smoothing_filter, SmoothingFilter)):
            raise TypeError(
                'The smoothing_filter argument must be None or an instance of '
                'SmoothingFilter! '
                f'Its current type is {classname(smoothing_filter)}!')
        if smoothing_filter is None:
            self.hist_smoothing_method = NoHistSmoothingMethod()
        else:
            self.hist_smoothing_method = NeighboringBinHistSmoothingMethod(
                (smoothing_filter.axis_kernel_array, UNSMOOTH_AXIS))

        if not isinstance(kde_smoothing, bool):
            raise ValueError(
                'The kde_smoothing argument must be an instance of bool! '
                f'Its current type is {classname(kde_smoothing)}!')

        # We have to figure out, which histogram bins are zero due to no
        # monte-carlo coverage, and which due to zero physics model
        # contribution.

        # Create a 2D histogram with only the MC events to determine the MC
        # coverage.
        (h, bins_logE, bins_sinDec) = np.histogram2d(
            data_logE,
            data_sinDec,
            bins=[
                logE_binning.binedges,
                sinDec_binning.binedges
            ],
            range=[
                logE_binning.range,
                sinDec_binning.range
            ],
            density=False)
        h = self._hist_smoothing_method.smooth(h)
        self._hist_mask_mc_covered = h > 0

        # Select the events which have MC coverage but zero physics
        # contribution, i.e. the physics model predicts zero contribution.
        mask = data_physicsweight == 0.

        # Create a 2D histogram with only the MC events that have zero physics
        # contribution. Note: By construction the zero physics contribution bins
        # are a subset of the MC covered bins.
        (h, bins_logE, bins_sinDec) = np.histogram2d(
            data_logE[mask],
            data_sinDec[mask],
            bins=[
                logE_binning.binedges,
                sinDec_binning.binedges
            ],
            range=[
                logE_binning.range,
                sinDec_binning.range
            ],
            density=False)
        h = self._hist_smoothing_method.smooth(h)
        self._hist_mask_mc_covered_zero_physics = h > 0

        if kde_smoothing:
            # If a bandwidth is passed, apply a KDE-based smoothing with the
            # given bandwidth parameter as bandwidth for the fit.
            kde_pdf_list = []
            data_logE_masked = data_logE[~mask]
            data_sinDec_masked = data_sinDec[~mask]

            for (sindec_lower, sindec_upper) in zip(
                    sinDec_binning.binedges[:-1],
                    sinDec_binning.binedges[1:]):

                sindec_mask = np.logical_and(
                    data_sinDec_masked >= sindec_lower,
                    data_sinDec_masked < sindec_upper
                )
                this_energy = data_logE_masked[sindec_mask]
                if sindec_lower >= 0:
                    kde = gaussian_kde(
                        this_energy,
                        bw_method=self._KDE_BW_NORTH)
                else:
                    kde = gaussian_kde(
                        this_energy,
                        bw_method=self._KDE_BW_SOUTH)
                kde_pdf_list.append(kde.evaluate(logE_binning.bincenters))

            h = np.vstack(kde_pdf_list).T

        else:
            # Create a 2D histogram with only the data which has physics
            # contribution. We will do the normalization along the logE
            # axis manually.
            data_weights = data_mcweight[~mask] * data_physicsweight[~mask]
            (h, bins_logE, bins_sinDec) = np.histogram2d(
                data_logE[~mask],
                data_sinDec[~mask],
                bins=[
                    logE_binning.binedges,
                    sinDec_binning.binedges
                ],
                weights=data_weights,
                range=[
                    logE_binning.range,
                    sinDec_binning.range
                ],
                density=False)

        # Calculate the normalization for each logE bin. Hence we need to sum
        # over the logE bins (axis 0) for each sin(dec) bin and need to divide
        # by the logE bin widths along the sin(dec) bins. The result array norm
        # is a 2D array of the same shape as h.
        norms = np.sum(h, axis=(0,))[np.newaxis, ...] * \
            np.diff(logE_binning.binedges)[..., np.newaxis]
        h /= norms
        h = self._hist_smoothing_method.smooth(h)

        self._hist_logE_sinDec = h

    @property
    def hist_smoothing_method(self):
        """The instance of HistSmoothingMethod defining the smoothing filter of
        the energy PDF histogram.
        """
        return self._hist_smoothing_method

    @hist_smoothing_method.setter
    def hist_smoothing_method(self, method):
        if not isinstance(method, HistSmoothingMethod):
            raise TypeError(
                'The hist_smoothing_method property must be an instance of '
                'HistSmoothingMethod! '
                f'Its current type is {classname(method)}!')
        self._hist_smoothing_method = method

    @property
    def hist(self):
        """(read-only) The 2D logE-sinDec histogram array.
        """
        return self._hist_logE_sinDec

    @property
    def hist_mask_mc_covered(self):
        """(read-only) The boolean ndarray holding the mask of the 2D histogram
        bins for which there is monte-carlo coverage.
        """
        return self._hist_mask_mc_covered

    @property
    def hist_mask_mc_covered_zero_physics(self):
        """(read-only) The boolean ndarray holding the mask of the 2D histogram
        bins for which there is monte-carlo coverage but zero physics
        contribution.
        """
        return self._hist_mask_mc_covered_zero_physics

    @property
    def hist_mask_mc_covered_with_physics(self):
        """(read-only) The boolean ndarray holding the mask of the 2D histogram
        bins for which there is monte-carlo coverage and has physics
        contribution.
        """
        return (
            self._hist_mask_mc_covered &
            ~self._hist_mask_mc_covered_zero_physics)

    def initialize_for_new_trial(
            self,
            tdm,
            tl=None,
            **kwargs):
        """Pre-compute the probability densitiy values of the trial data,
        which has to be done only once for a particular trial data.
        """

        logE_binning = self.get_binning('log_energy')
        sinDec_binning = self.get_binning('sin_dec')

        logE_idx = np.digitize(
            tdm['log_energy'], logE_binning.binedges) - 1
        sinDec_idx = np.digitize(
            tdm['sin_dec'], sinDec_binning.binedges) - 1

        with TaskTimer(tl, 'Evaluating logE-sinDec histogram.'):
            self._pd = self._hist_logE_sinDec[(logE_idx, sinDec_idx)]

    def assert_is_valid_for_trial_data(
            self,
            tdm,
            tl=None):
        """Checks if this energy PDF covers the entire value range of the trail
        data events.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial data events.
            The following data fields need to exist:

                log_energy : float
                    The base-10 logarithm of the reconstructed energy value.
                sin_dec : float
                    The sine of the declination value of the event.
        tl : instance of TimeLord | None
            The optional instance of TimeLord to measure timing information.

        Raises
        ------
        ValueError
            If parts of the trial data is outside the value range of this
            PDF.
        """
        log10emu = tdm['log_energy']
        log10emu_axis = self.axes['log_energy']
        if np.min(log10emu) < log10emu_axis.vmin:
            raise ValueError(
                f'The minimum log10emu value {np.min(log10emu):g} of the trial '
                'data is lower than the minimum value of the PDF '
                f'{log10emu_axis.vmin:g}!')
        if np.max(log10emu) > log10emu_axis.vmax:
            raise ValueError(
                f'The maximum log10emu value {np.max(log10emu):g} of the trial '
                'data is larger than the maximum value of the PDF '
                f'{log10emu_axis.vmax}:g!')

        sindecmu = tdm['sin_dec']
        sindecmu_axis = self.axes['sin_dec']
        if np.min(sindecmu) < sindecmu_axis.vmin:
            raise ValueError(
                f'The minimum sindecmu value {np.min(sindecmu):g} of the trial '
                'data is lower than the minimum value of the PDF '
                f'{sindecmu_axis.vmin:g}!')
        if np.max(sindecmu) > sindecmu_axis.vmax:
            raise ValueError(
                f'The maximum sindecmu value {np.max(sindecmu):g} of the trial '
                'data is larger than the maximum value of the PDF '
                f'{sindecmu_axis.vmax:g}!')

    def get_pd(
            self,
            tdm,
            params_recarray=None,
            tl=None):
        """Calculates the energy probability density (in 1/log10(E/GeV)) of each
        trial data event.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the data events for which the
            probability should be calculated for. The following data fields must
            exist:

            log_energy : float
                The base-10 logarithm of the energy value of the event.
            sin_dec : float
                The sin(declination) value of the event.

        params_recarray : None
            Unused interface parameter.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        pd : instance of ndarray
            The (N_selected_events,)-shaped numpy ndarray holding the energy
            probability density value for each trial data event.
        grads : dict
            The dictionary holding the gradients of the probability density
            w.r.t. each global fit parameter. By definition this PDF does not
            depend on any fit parameter, hence, this dictionary is empty.
        """
        grads = dict()

        return (self._pd, grads)


class PDDataBackgroundI3EnergyPDF(
        PDBackgroundI3EnergyPDF):
    """This is the IceCube energy background PDF, which gets constructed from
    the experimental data of the public data.
    """

    def __init__(
            self,
            data_exp,
            logE_binning,
            sinDec_binning,
            smoothing_filter=None,
            kde_smoothing=False,
            **kwargs):
        """Constructs a new IceCube energy background PDF from experimental
        data.

        Parameters
        ----------
        data_exp : instance of DataFieldRecordArray
            The array holding the experimental data. The following data fields
            must exist:

            log_energy : float
                The base-10 logarithm of the reconstructed energy value of the
                data event.
            sin_dec : float
                The sine of the reconstructed declination of the data event.

        logE_binning : instance of BinningDefinition
            The binning definition for the binning in log10(E).
        sinDec_binning : instance of BinningDefinition
            The binning definition for the sin(declination).
        smoothing_filter : instance of SmoothingFilter | None
            The smoothing filter to use for smoothing the energy histogram.
            If None, no smoothing will be applied.
        kde_smoothing : bool
            Apply a kde smoothing to the energy pdf for each bin in sin(dec).
            This is useful for signal injections, because it ensures that the
            background is not zero when injecting high energy events.
            Default: False.
        """
        if not isinstance(data_exp, DataFieldRecordArray):
            raise TypeError(
                'The data_exp argument must be an instance of '
                'DataFieldRecordArray! '
                f'Its current type is {classname(data_exp)}!')

        data_logE = data_exp['log_energy']
        data_sinDec = data_exp['sin_dec']
        # For experimental data, the MC and physics weight are unity.
        data_mcweight = np.ones((len(data_exp),))
        data_physicsweight = data_mcweight

        # Create the energy PDF using the base class.
        super().__init__(
            data_logE=data_logE,
            data_sinDec=data_sinDec,
            data_mcweight=data_mcweight,
            data_physicsweight=data_physicsweight,
            logE_binning=logE_binning,
            sinDec_binning=sinDec_binning,
            smoothing_filter=smoothing_filter,
            kde_smoothing=kde_smoothing,
            **kwargs)


class PDMCBackgroundI3EnergyPDF(
        EnergyPDF,
        IsBackgroundPDF,
        UsesBinning):
    """This class provides a background energy PDF constructed from the public
    data and a monte-carlo background flux model.
    """

    def __init__(
            self,
            pdf_log10emu_sindecmu,
            log10emu_binning,
            sindecmu_binning,
            **kwargs):
        """Constructs a new background energy PDF with the given PDF data and
        binning.

        Parameters
        ----------
        pdf_log10emu_sindecmu : instance of numpy ndarray
            The (n_log10emu, n_sindecmu)-shaped 2D numpy ndarray holding the
            PDF values in unit 1/log10(E_mu/GeV).
            A copy of this data will be created and held within this class
            instance.
        log10emu_binning : instance of BinningDefinition
            The binning definition for the binning in log10(E_mu/GeV).
        sindecmu_binning : instance of BinningDefinition
            The binning definition for the binning in sin(dec_mu).
        """
        if not isinstance(pdf_log10emu_sindecmu, np.ndarray):
            raise TypeError(
                'The pdf_log10emu_sindecmu argument must be an instance of '
                'numpy.ndarray!')
        if not isinstance(sindecmu_binning, BinningDefinition):
            raise TypeError(
                'The sindecmu_binning argument must be an instance of '
                'BinningDefinition!')
        if not isinstance(log10emu_binning, BinningDefinition):
            raise TypeError(
                'The log10emu_binning argument must be an instance of '
                'BinningDefinition!')

        super().__init__(
            pmm=None,
            **kwargs)

        self.add_axis(PDFAxis(
            log10emu_binning.name,
            log10emu_binning.lower_edge,
            log10emu_binning.upper_edge,
        ))

        self.add_axis(PDFAxis(
            sindecmu_binning.name,
            sindecmu_binning.lower_edge,
            sindecmu_binning.upper_edge,
        ))

        self._hist_logE_sinDec = np.copy(pdf_log10emu_sindecmu)
        self.add_binning(log10emu_binning, name='log_energy')
        self.add_binning(sindecmu_binning, name='sin_dec')

    def assert_is_valid_for_trial_data(
            self,
            tdm,
            tl=None):
        """Checks if this energy PDF covers the entire value range of the trail
        data events.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial data events.
            The following data fields need to exist:

                log_energy : float
                    The base-10 logarithm of the reconstructed energy value.
                sin_dec : float
                    The sine of the declination value of the event.
        tl : instance of TimeLord | None
            The optional instance of TimeLord to measure timing information.

        Raises
        ------
        ValueError
            If parts of the trial data is outside the value range of this
            PDF.
        """
        log10emu = tdm['log_energy']
        log10emu_axis = self.get_axis(0)
        if np.min(log10emu) < log10emu_axis.vmin:
            raise ValueError(
                f'The minimum log10emu value {np.min(log10emu):g} of the trial '
                'data is lower than the minimum value of the PDF '
                f'{log10emu_axis.vmin:g}!')
        if np.max(log10emu) > log10emu_axis.vmax:
            raise ValueError(
                f'The maximum log10emu value {np.max(log10emu):g} of the trial '
                'data is larger than the maximum value of the PDF '
                f'{log10emu_axis.vmax}:g!')

        sindecmu = tdm['sin_dec']
        sindecmu_axis = self.get_axis(1)
        if np.min(sindecmu) < sindecmu_axis.vmin:
            raise ValueError(
                f'The minimum sindecmu value {np.min(sindecmu):g} of the trial '
                'data is lower than the minimum value of the PDF '
                f'{sindecmu_axis.vmin:g}!')
        if np.max(sindecmu) > sindecmu_axis.vmax:
            raise ValueError(
                f'The maximum sindecmu value {np.max(sindecmu):g} of the trial '
                'data is larger than the maximum value of the PDF '
                f'{sindecmu_axis.vmax:g}!')

    def get_pd(
            self,
            tdm,
            params_recarray=None,
            tl=None):
        """Gets the probability density for the given trial data events.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial data events.
            The following data fields need to exist:

                log_energy : float
                    The base-10 logarithm of the reconstructed energy value.
                sin_dec : float
                    The sine of the declination value of the event.

        params_recarray : None
            Unused interface argument.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to measure
            timing information.

        Returns
        -------
        pd : instance of ndarray
            The (N_selected_events,)-shaped numpy ndarray holding the
            probability density value for each event.
        grads : dict
            The dictionary holding the gradients of the probability density
            w.r.t. each global fit parameter. By definition this PDF does not
            depend on any fit parameter, hence, this dictionary is empty.
        """
        log10emu = tdm['log_energy']
        sindecmu = tdm['sin_dec']

        log10emu_idxs = np.digitize(
            log10emu, self.get_binning('log_energy').binedges) - 1
        sindecmu_idxs = np.digitize(
            sindecmu, self.get_binning('sin_dec').binedges) - 1

        with TaskTimer(tl, 'Evaluating sindecmu-log10emu PDF.'):
            pd = self._hist_logE_sinDec[(log10emu_idxs, sindecmu_idxs)]

        grads = dict()

        return (pd, grads)



# -*- coding: utf-8 -*-

import numpy as np

from scipy import integrate

from skyllh.analyses.i3.publicdata_ps.aeff import (
    PDAeff,
)
from skyllh.analyses.i3.publicdata_ps.utils import (
    FctSpline1D,
)
from skyllh.analyses.i3.publicdata_ps.smearing_matrix import (
    PDSmearingMatrix,
)
from skyllh.core.binning import (
    get_bincenters_from_binedges,
)
from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core.flux_model import (
    FactorizedFluxModel,
)
from skyllh.core.multiproc import (
    IsParallelizable,
    parallelize,
)
from skyllh.core.parameters import (
    ParameterGrid,
    ParameterGridSet,
)
from skyllh.core.pdf import (
    IsSignalPDF,
    PDF,
    PDFAxis,
    PDFSet,
)
from skyllh.core.py import (
    classname,
    module_classname,
)
from skyllh.core.timing import (
    TaskTimer,
)
from skyllh.i3.dataset import (
    I3Dataset,
)


class PDSignalEnergyPDF(
        PDF,
        IsSignalPDF,
):
    """This class provides a signal energy PDF for a spectrial index value.
    """
    def __init__(
            self,
            f_e_spl,
            **kwargs,
    ):
        """Creates a new signal energy PDF instance for a particular spectral
        index value.

        Parameters
        ----------
        f_e_spl : instance of FctSpline1D
            The instance of FctSpline1D representing the spline of the energy
            PDF.
        """
        super().__init__(
            pmm=None,
            **kwargs)

        if not isinstance(f_e_spl, FctSpline1D):
            raise TypeError(
                'The f_e_spl argument must be an instance of FctSpline1D! '
                f'Its current type is {classname(f_e_spl)}!')

        self.f_e_spl = f_e_spl

        self.log10_reco_e_lower_binedges = self.f_e_spl.x_binedges[:-1]
        self.log10_reco_e_upper_binedges = self.f_e_spl.x_binedges[1:]

        self.log10_reco_e_min = self.log10_reco_e_lower_binedges[0]
        self.log10_reco_e_max = self.log10_reco_e_upper_binedges[-1]

        # Add the PDF axes.
        self.add_axis(
            PDFAxis(
                name='log_energy',
                vmin=self.log10_reco_e_min,
                vmax=self.log10_reco_e_max))

        # Check integrity.
        integral = integrate.quad(
            self.f_e_spl.evaluate,
            self.log10_reco_e_min,
            self.log10_reco_e_max,
            limit=200,
            full_output=1
        )[0] / self.f_e_spl.norm
        if not np.isclose(integral, 1):
            raise ValueError(
                'The integral over log10_reco_e of the energy term must be '
                f'unity! But it is {integral}!')

    def assert_is_valid_for_trial_data(
            self,
            tdm,
            tl=None):
        pass

    def get_pd_by_log10_reco_e(
            self,
            log10_reco_e,
            tl=None):
        """Calculates the probability density for the given log10(E_reco/GeV)
        values using the spline representation of the PDF.

        Parameters
        ----------
        log10_reco_e : instance of ndarray
            The (n_log10_reco_e,)-shaped numpy ndarray holding the
            log10(E_reco/GeV) values for which the energy PDF should get
            evaluated.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to measure
            timing information.

        Returns
        -------
        pd : instance of numpy ndarray
            The (n_log10_reco_e,)-shaped numpy ndarray with the probability
            density for each energy value.
        """
        # Select energy values that actually have a signal energy PDF.
        # All other energy values will get zero signal probability density.
        m = (
            (log10_reco_e >= self.log10_reco_e_min) &
            (log10_reco_e < self.log10_reco_e_max)
        )

        with TaskTimer(tl, 'Evaluate PDSignalEnergyPDF'):
            pd = np.zeros((len(log10_reco_e),), dtype=np.double)
            pd[m] = self.f_e_spl(log10_reco_e[m]) / self.f_e_spl.norm

        return pd

    def get_pd(
            self,
            tdm,
            params_recarray=None,
            tl=None):
        """Calculates the probability density for all given trial data events
        and sources.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial data events for
            which the probability density should be looked up.
            The following data fields must be present:

            log_energy : float
                The base-10 logarithm of the reconstructed energy.

        params_recarray : None
            Unused interface argument.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to measure
            timing information.

        Returns
        -------
        pd : instance of ndarray
            The (N_values,)-shaped numpy ndarray holding the probability density
            for each trial data event and source.
        grads : dict
            The dictionary holding the gradient values for each global fit
            parameter. By definition this PDF does not depend on any fit
            parameters, hence, this is an empty dictionary.
        """
        evt_idxs = tdm.src_evt_idxs[1]

        log10_reco_e = np.take(tdm['log_energy'], evt_idxs)

        pd = self.get_pd_by_log10_reco_e(
            log10_reco_e=log10_reco_e,
            tl=tl)

        grads = dict()

        return (pd, grads)


class PDSignalEnergyPDFSet(
        PDFSet,
        IsSignalPDF,
        PDF,
        IsParallelizable,
):
    """This class provides a signal energy PDF set using the public data.
    It creates a set of PDSignalEnergyPDF instances, one for each spectral
    index value on a grid.
    """
    def __init__(
            self,
            ds,
            src_dec,
            fluxmodel,
            param_grid_set,
            ncpu=None,
            ppbar=None,
            **kwargs,
    ):
        """Creates a new PDSignalEnergyPDFSet instance for the public data.

        Parameters
        ----------
        ds : instance of Dataset
            The instance of Dataset that defines the dataset of the public data.
        src_dec : float
            The declination of the source in radians.
        fluxmodel : instance of FactorizedFluxModel
            The instance of FactorizedFluxModel that defines the source's flux
            model.
        param_grid_set : instance of ParameterGrid | instance of ParameterGridSet
            The parameter grid set defining the grids of the parameters this
            energy PDF set depends on.
        ncpu : int | None
            The number of CPUs to utilize. Global setting will take place if
            not specified, i.e. set to None.
        ppbar : instance of ProgressBar | None
            The instance of ProgressBar for the optional parent progress bar.
        """
        self._logger = get_logger(module_classname(self))

        # Check for the correct types of the arguments.
        if not isinstance(ds, I3Dataset):
            raise TypeError(
                'The ds argument must be an instance of I3Dataset!')

        if not isinstance(fluxmodel, FactorizedFluxModel):
            raise TypeError(
                'The fluxmodel argument must be an instance of '
                'FactorizedFluxModel! '
                f'Its current type is {classname(fluxmodel)}')

        if (not isinstance(param_grid_set, ParameterGrid)) and\
           (not isinstance(param_grid_set, ParameterGridSet)):
            raise TypeError(
                'The param_grid_set argument must be an instance of type '
                'ParameterGrid or ParameterGridSet! '
                f'Its current type is {classname(param_grid_set)}!')

        # Extend the param_grid_set to allow for parameter interpolation
        # values at the grid edges.
        param_grid_set = param_grid_set.copy()
        param_grid_set.add_extra_lower_and_upper_bin()

        super().__init__(
            param_grid_set=param_grid_set,
            ncpu=ncpu,
            **kwargs)

        # Load the smearing matrix.
        sm = PDSmearingMatrix(
            pathfilenames=ds.get_abs_pathfilename_list(
                ds.get_aux_data_definition('smearing_datafile')))

        # Select the slice of the smearing matrix corresponding to the
        # source declination band.
        true_dec_idx = sm.get_true_dec_idx(src_dec)
        sm_pdf = sm.pdf[:, true_dec_idx]

        # Only look at true neutrino energies for which a recostructed
        # muon energy distribution exists in the smearing matrix.
        (min_log_true_e,
         max_log_true_e) = sm.get_true_log_e_range_with_valid_log_e_pdfs(
            true_dec_idx)
        log_true_e_mask = np.logical_and(
            sm.log10_true_enu_binedges >= min_log_true_e,
            sm.log10_true_enu_binedges <= max_log_true_e)
        true_enu_binedges = np.power(
            10, sm.log10_true_enu_binedges[log_true_e_mask])
        true_enu_binedges_lower = true_enu_binedges[:-1]
        true_enu_binedges_upper = true_enu_binedges[1:]
        valid_true_e_idxs = [
            sm.get_log10_true_e_idx(0.5 * (he + le))
            for (he, le) in zip(
                sm.log10_true_enu_binedges[log_true_e_mask][1:],
                sm.log10_true_enu_binedges[log_true_e_mask][:-1])
        ]

        xvals_binedges = ds.get_binning_definition('log_energy').binedges
        xvals = get_bincenters_from_binedges(xvals_binedges)

        # Calculate the neutrino enegry bin widths in GeV.
        d_enu = np.diff(true_enu_binedges)
        self._logger.debug(
            'dE_nu = {}'.format(d_enu)
        )

        # Load the effective area.
        aeff = PDAeff(
            pathfilenames=ds.get_abs_pathfilename_list(
                ds.get_aux_data_definition('eff_area_datafile')))

        # Calculate the probability to detect a neutrino of energy
        # E_nu given a neutrino declination: p(E_nu|dec).
        det_prob = aeff.get_detection_prob_for_decnu(
            decnu=src_dec,
            enu_min=true_enu_binedges[:-1],
            enu_max=true_enu_binedges[1:],
            enu_range_min=true_enu_binedges[0],
            enu_range_max=true_enu_binedges[-1]
        )

        self._logger.debug('det_prob = {}, sum = {}'.format(
            det_prob, np.sum(det_prob)))

        if not np.isclose(np.sum(det_prob), 1):
            self._logger.warn(
                'The sum of the detection probabilities is not unity! It is '
                '{}.'.format(np.sum(det_prob)))

        psi_edges_bw = sm.psi_upper_edges - sm.psi_lower_edges
        ang_err_bw = sm.ang_err_upper_edges - sm.ang_err_lower_edges

        # Create the energy pdf for different gamma values.
        def create_energy_pdf(sm_pdf, fluxmodel, gridparams):
            """Creates an energy pdf for a specific gamma value.
            """
            # Create a copy of the FluxModel with the given flux parameters.
            # The copy is needed to not interfer with other CPU processes.
            my_fluxmodel = fluxmodel.copy(newparams=gridparams)

            self._logger.debug(
                f'Generate signal energy PDF for parameters {gridparams} in '
                f'{len(valid_true_e_idxs)} E_nu bins.')

            # Calculate the flux probability p(E_nu|gamma).
            flux_prob = (
                my_fluxmodel.energy_profile.get_integral(
                    true_enu_binedges_lower,
                    true_enu_binedges_upper
                ) /
                my_fluxmodel.energy_profile.get_integral(
                    true_enu_binedges[0],
                    true_enu_binedges[-1]
                )
            )
            if not np.isclose(np.sum(flux_prob), 1):
                self._logger.warn(
                    'The sum of the flux probabilities is not unity! It is '
                    '{}.'.format(np.sum(flux_prob)))

            self._logger.debug(
                'flux_prob = {}, sum = {}'.format(
                    flux_prob, np.sum(flux_prob))
            )

            p = flux_prob * det_prob

            true_e_prob = p / np.sum(p)

            self._logger.debug(
                'true_e_prob = {}'.format(
                    true_e_prob))

            def create_reco_e_pdf_for_true_e(idx, true_e_idx):
                """This functions creates a spline for the reco energy
                distribution given a true neutrino engery.
                """
                # Create the energy PDF f_e = P(log10_E_reco|dec) =
                # \int dPsi dang_err P(E_reco,Psi,ang_err).
                f_e = np.sum(
                    sm_pdf[true_e_idx] *
                    psi_edges_bw[true_e_idx, true_dec_idx, :, :, np.newaxis] *
                    ang_err_bw[true_e_idx, true_dec_idx, :, :, :],
                    axis=(-1, -2)
                )

                # Build the spline for this P(E_reco|E_nu). Weigh the pdf
                # with the true neutrino energy probability (flux prob).
                log10_reco_e_binedges = sm.log10_reco_e_binedges[
                    true_e_idx, true_dec_idx]

                p = f_e * true_e_prob[idx]

                spline = FctSpline1D(p, log10_reco_e_binedges)

                return spline(xvals)

            # Integrate over the true neutrino energy and spline the output.
            sum_pdf = np.sum(
                [
                    create_reco_e_pdf_for_true_e(i, true_e_idx)
                    for (i, true_e_idx) in enumerate(valid_true_e_idxs)
                ],
                axis=0)

            spline = FctSpline1D(sum_pdf, xvals_binedges, norm=True)

            pdf = PDSignalEnergyPDF(spline, cfg=self._cfg)

            return pdf

        args_list = [
            ((sm_pdf, fluxmodel, gridparams), {})
            for gridparams in self.gridparams_list
        ]

        pdf_list = parallelize(
            create_energy_pdf,
            args_list,
            ncpu=self.ncpu,
            ppbar=ppbar)

        del sm_pdf

        # Save all the energy PDF objects in the PDFSet PDF registry with
        # the hash of the individual parameters as key.
        for (gridparams, pdf) in zip(self.gridparams_list, pdf_list):
            self.add_pdf(pdf, gridparams)



# -*- coding: utf-8 -*-

"""This file defines IceCube specific global configuration.
"""

from skyllh.core.datafields import (
    DataFieldStages as DFS,
)


def add_icecube_specific_analysis_required_data_fields(cfg):
    """Adds IceCube specific data fields required by an IceCube analysis to
    the given local configuration.

    Parameters
    ----------
    cfg : instance of Config
        The instance of Config holding the local configuration.
    """
    cfg['datafields']['azi'] = DFS.ANALYSIS_EXP
    cfg['datafields']['zen'] = DFS.ANALYSIS_EXP
    cfg['datafields']['sin_dec'] = DFS.ANALYSIS_EXP
    cfg['datafields']['sin_true_dec'] = DFS.ANALYSIS_MC
